/**
	$Id: ExperimentExplorerPlotServlet.java 5319 2010-04-20 08:02:16Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.servlet;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.font.FontRenderContext;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.jfree.chart.ChartRenderingInfo;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.CategoryLabelPositions;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.entity.CategoryLabelEntity;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.EntityCollection;
import org.jfree.chart.entity.StandardEntityCollection;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.renderer.category.BarRenderer;
import org.jfree.chart.renderer.category.BoxAndWhiskerRenderer;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.chart.renderer.category.StandardBarPainter;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.data.statistics.BoxAndWhiskerItem;
import org.jfree.data.statistics.DefaultBoxAndWhiskerCategoryDataset;

import net.sf.basedb.clients.web.ExperimentExplorer;
import net.sf.basedb.clients.web.ExperimentExplorer.AnnotationGroup;
import net.sf.basedb.clients.web.ExperimentExplorer.AnnotationSummary;
import net.sf.basedb.clients.web.taglib.table.TableColumn;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.ThrowableUtil;


/**
	Plot servlet for generating plots from experiment explorer.
	
	@author Nicklas
	@version 2.14
	@base.modified $Date: 2010-04-20 10:02:16 +0200 (Tue, 20 Apr 2010) $
*/
public class ExperimentExplorerPlotServlet
	extends HttpServlet
{
	
	private static final long serialVersionUID = -445763412868957935L;
	private int maxWidth = 1000;
	private int maxHeight = 800;
	private int defaultWidth = 600;
	private int defaultHeight = 400;
	private String defaultFormat = "png";

	@Override
	public void init()
		throws ServletException
	{
		ServletConfig cfg = getServletConfig();
		maxWidth = Values.getInt(cfg.getInitParameter("maxWidth"), maxWidth);
		maxHeight = Values.getInt(cfg.getInitParameter("maxHeight"), maxHeight);
		defaultWidth = Values.getInt(cfg.getInitParameter("defaultWidth"), defaultWidth);
		defaultHeight = Values.getInt(cfg.getInitParameter("defaultHeight"), defaultHeight);
		defaultFormat = Values.getString(cfg.getInitParameter("defaultFormat"), defaultFormat);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void doGet(HttpServletRequest request, HttpServletResponse response)
		throws IOException, ServletException
	{
		final String ID = request.getParameter("ID");
		
		// Where to get the source data
		final int bioassaySetId = Values.getInt(request.getParameter("bioAssaySetId"));
		final int reporterIndex = Values.getInt(request.getParameter("reporterIndex"));
		final int positionIndex = Values.getInt(request.getParameter("positionIndex"));
		final String plotType = Values.getString(request.getParameter("type"), "assay");
		final String subType = Values.getString(request.getParameter("subtype"), "line");
		
		// Where to deliver the image
		final boolean download = Values.getBoolean(request.getParameter("download"));
		final String saveAs = Values.getStringOrNull(request.getParameter("saveAs"));
		final boolean overwrite = Values.getBoolean(request.getParameter("overwrite"));
		
		// Image settings, width, height and format
		int width = Values.getInt(request.getParameter("width"), defaultWidth);
		if (width <= 100) width = defaultWidth;
		if (width > maxWidth) width = maxWidth;
		int height = Values.getInt(request.getParameter("height"), defaultHeight);
		if (height <= 100) height = defaultHeight;
		if (height > maxHeight) height = maxHeight;
		String format = Values.getString(request.getParameter("format"), defaultFormat);
		if (!format.equals("jpeg") && !format.equals("png"))
		{
			format = defaultFormat;
		}

		// Formulas to plot
		final boolean yLog = Values.getBoolean(request.getParameter("yLog"));
		final String yFormula = request.getParameter("y");
		Formula.AverageMethod averageMethod = null;
		if (positionIndex == ExperimentExplorer.SPOT_AVG)
		{
			String avg = Values.getString(request.getParameter("averageMethod"), 
					"ARITHMETIC_MEAN");
			averageMethod = Formula.AverageMethod.valueOf(avg);
		}
		
		// Axis, legends and titles
		final String title = Values.getStringOrNull(request.getParameter("title"));
		final String subTitle = Values.getStringOrNull(request.getParameter("subTitle"));
		String yLabel = Values.getStringOrNull(request.getParameter("yLabel"));
		if (yLog && yLabel != null) yLabel = "log2("+yLabel+")";
		final boolean showXLabels = Values.getBoolean(request.getParameter("showXLabels"));
		final boolean showAnnotations = true;

		// Box plot specific parameters
		final int annotationTypeId = Values.getInt(request.getParameter("annotationTypeId"));
		
		final SessionControl sc = Application.getSessionControl(ID, request.getRemoteAddr());
		DbControl dc = null;
		BufferedImage image = null;
		// Outer try-finally is for making sure objects (eg. DbControl) are closed properly
		try
		{
			// Generate the image
			try
			{
				dc = sc.newDbControl();
				final BioAssaySet bioAssaySet = BioAssaySet.getById(dc, bioassaySetId);
				final ExperimentExplorer explorer = ExperimentExplorer.getExplorer(bioAssaySet);
			
				
				// Create some common items used for all plot types
		      NumberAxis valueAxis = new NumberAxis(yLabel);
		      CategoryAxis categoryAxis = new CategoryAxis();

				// Create the query
				DynamicSpotQuery query = explorer.getSpotQuery(dc, reporterIndex, positionIndex, new ArrayList<TableColumn>());
				query.setDisableRawJoinOnMergedData(true);

		      // Convert formulas to Expression:s
				Expression y = yFormula == null ? null : 
					BioAssaySetUtil.createJepExpression(dc, yFormula, query);
				if (yLog) y = Expressions.log2(y);
				if (averageMethod != null) y = averageMethod.getAverageExpression(y);

				query.select(Selects.expression(y, "y"));

		      String[] categoryAnnotations = null;
		      
				JFreeChart chart = null;
				if ("annotation".equals(plotType))
				{
					// Collect all spot data values in one list per annotation group
					Map<AnnotationGroup, List<Float>> valuesPerGroup = new HashMap<AnnotationGroup, List<Float>>();
					AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
					AnnotationSummary summary = explorer.getAnnotationSummary(dc, at);
					DynamicResultIterator it = query.iterate(dc);
					int valueIndex = it.getIndex("y");
					int columnIndex = it.getIndex(VirtualColumn.COLUMN.getName());
					int posIndex = positionIndex == ExperimentExplorer.SPOT_ALL ? 
						it.getIndex(VirtualColumn.POSITION.getName()) : 0;
					while (it.hasNext())
					{
						SqlResult result = it.next();
						Float value = result.getFloat(valueIndex);
						if (value == null || value.isNaN()) continue;
						short column = result.getShort(columnIndex);
						AnnotationGroup grp = summary.getAnnotationGroup(column);
						List<Float> values = valuesPerGroup.get(grp);
						if (values == null)
						{
							values = new ArrayList<Float>();
							valuesPerGroup.put(grp, values);
						}
						values.add(value);
					}
					
					// Create the JFreeChart box-and-whisker dataset
					DefaultBoxAndWhiskerCategoryDataset dataset = new DefaultBoxAndWhiskerCategoryDataset();
					for (AnnotationGroup grp : summary.getAnnotationGroups())
					{
						List<Float> values = valuesPerGroup.get(grp);
						if (values == null || values.size() == 0) continue;
						Collections.sort(values);
						BoxAndWhiskerItem box = calculateBoxAndWhiskerStatistics(values);
						dataset.add(box, "", new ColumnKey(grp.hashCode(), grp.getTitle()));
					}
					categoryAxis.setLabel(at.getName());
					BoxAndWhiskerRenderer renderer = new BoxAndWhiskerRenderer();
					renderer.setMeanVisible(false);
					renderer.setMedianVisible(true);
					renderer.setFillBox(false);
					renderer.setAutoPopulateSeriesPaint(false);
					renderer.setBasePaint(Color.BLACK);
					renderer.setBaseOutlinePaint(Color.BLACK);
					Plot plot = new CategoryPlot(dataset, categoryAxis, valueAxis, renderer);
					chart = new JFreeChart(plot);
				}
				else if ("assay".equals(plotType))
				{
					// Load annotation summaries
					Set<Integer> annotationTypes = explorer.getAnnotationTypeIds();
					List<AnnotationSummary> summaries = new ArrayList<AnnotationSummary>(3);
					if (showAnnotations && annotationTypes != null && annotationTypes.size() > 0)
					{
						categoryAnnotations = new String[Math.min(3, annotationTypes.size())];
						int i = 0;
						for (Integer atId : annotationTypes)
						{
							AnnotationType at = AnnotationType.getById(dc, atId);
							summaries.add(explorer.getAnnotationSummary(dc, at));
							categoryAnnotations[i] = at.getName();
							++i;
							if (summaries.size() == 3) break;
						}
					}

					// Load bioassay information and related annotation group summaries
					Map<Short, ColumnKey> columnToLabel = new HashMap<Short, ColumnKey>();
					for (BioAssay ba : bioAssaySet.getBioAssays().list(dc))
					{
						short column = ba.getDataCubeColumnNo();
						ColumnKey key = new ColumnKey(ba.getId(), showXLabels ? ba.getName(): " ");
						for (AnnotationSummary summary : summaries)
						{
							AnnotationGroup grp = summary.getAnnotationGroup(column);
							if (grp != null) key.setColor(summary.getAnnotationType().getName(), grp.getAwtColor());
						}
						columnToLabel.put(column, key);
					}

					// Load the spot data into the dataset
					DefaultCategoryDataset dataset = new DefaultCategoryDataset();
					DynamicResultIterator it = query.iterate(dc);
					int valueIndex = it.getIndex("y");
					int columnIndex = it.getIndex(VirtualColumn.COLUMN.getName());
					int posIndex = positionIndex == ExperimentExplorer.SPOT_ALL ? 
						it.getIndex(VirtualColumn.POSITION.getName()) : 0;
					while (it.hasNext())
					{
						SqlResult result = it.next();
						Float value = result.getFloat(valueIndex);
						if (value == null || value.isNaN()) continue;
						short column = result.getShort(columnIndex);
						int position = posIndex > 0 ? result.getInt(posIndex) : 0;
						dataset.addValue((Number)value, position, columnToLabel.get(column));
					}
				
					// Create the plot and chart objects
			      categoryAxis.setCategoryLabelPositions(CategoryLabelPositions.UP_90);
			      CategoryItemRenderer renderer = null;
			      if ("line".equals(subType))
			      {
		      		renderer = new LineAndShapeRenderer();
			      }
			      else
			      {
		      		BarRenderer br = new BarRenderer();
		      		renderer = br;
		      		br.setShadowVisible(false);
		      		br.setBarPainter(new StandardBarPainter());
		      		br.setDrawBarOutline(true);
			      }
					Plot plot = new CategoryPlot(dataset, categoryAxis, valueAxis, renderer);
					chart = new JFreeChart(plot);
				}
				chart.removeLegend();
				if (title != null) chart.setTitle(title);
				if (subTitle != null) chart.addSubtitle(new TextTitle(subTitle));

				// Render the image
				EntityCollection chartEntities = null;
				ChartRenderingInfo chartInfo = null;
				int chartHeight = height;
				if (categoryAnnotations != null)
				{
					chartEntities = new StandardEntityCollection();
					chartInfo = new ChartRenderingInfo(chartEntities);
					chartHeight -= 15 * categoryAnnotations.length;
				}
				BufferedImage temp = chart.createBufferedImage(width, chartHeight, chartInfo);
				if (categoryAnnotations == null)
				{
					// Use the image that JFreeChart generated if we don't have any annotations
					image = temp;
				}
				else
				{
					// Otherwise... create a new (bigger) image...
					image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
					Graphics2D g = image.createGraphics();
					g.setBackground(Color.WHITE);
					g.clearRect(0, 0, width, height);
					g.setColor(Color.BLACK);
					// ... and copy the image from JFreeChart into it
					g.drawImage(temp, 0, 0, width, chartHeight, 0, 0, width, chartHeight, null);
					
					// Draw annotation color rectangles
					int minX = width;
					for (ChartEntity chartEntity : (Collection<ChartEntity>)chartEntities.getEntities())
					{
						if (chartEntity.getClass() == CategoryLabelEntity.class)
						{
							// This is the category (x) axis label object
							CategoryLabelEntity lbl = (CategoryLabelEntity)chartEntity;
							// Get the coordinates for the label
							Rectangle lblCoordinates = chartEntity.getArea().getBounds();
							if (lblCoordinates.getMinX() < minX) minX = (int)lblCoordinates.getMinX();
							
							ColumnKey key = (ColumnKey)lbl.getKey();
							int currentY = chartHeight;
							for (String annotation : categoryAnnotations)
							{
								Color color = key.getColor(annotation);
								if (color != null)
								{
									// Draw colored rectangles below the label
									Rectangle r = new Rectangle((int)lblCoordinates.getMinX(), currentY, (int)lblCoordinates.getWidth(), 10);
									g.setColor(color);
									g.fill(r);
									g.setColor(Color.LIGHT_GRAY);
									g.draw(r);
								}
								currentY += 15;
							}
						}
					}
					// Draw annotation titles
					int currentY = chartHeight + 10;
					g.setPaint(categoryAxis.getLabelPaint());
					g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
					Font lblFont = categoryAxis.getLabelFont();
					FontRenderContext context = g.getFontRenderContext();

					for (String annotation : categoryAnnotations)
					{
						// Try to calculate coordinate so that the text is right-aligned 
						// near the first label
						Rectangle2D bounds = lblFont.getStringBounds(annotation, context);
						int x = minX - (int)bounds.getWidth() - 5;
						if (x < 0) 
						{
							x = 0;
							annotation = StringUtil.trimString(annotation, (int)(annotation.length() * minX / bounds.getWidth()));
						}
						g.drawString(annotation, x, currentY);
						currentY += 15;
					}
				}
			}
			// Catch all problems that happens when trying to generate the image
			catch (Throwable t)
			{
				t.printStackTrace();
				if (saveAs == null)
				{
					// Generate error image
					try
					{
						image = ThrowableUtil.stackTraceToImage(t, width, height);
					}
					catch (Throwable t2)
					{
						t2.printStackTrace();
					}
				}
				else 
				{
					// Let servlet container redirect to appropriate error page
					throw new ServletException(t);
				}
			}
			
			// Deliver the image
			if (saveAs != null)
			{
				// The image should be saved in the BASE file system
				// and we response with a regular redirect with a message
				saveImageToFile(dc, image, saveAs, format, overwrite);
				dc.commit();
				response.sendRedirect(request.getContextPath()+
						"/common/close_popup.jsp?message=Plot+saved+successfully");
			}
			else
			{
				// The image should be sent in the response
				if (download)
				{
					response.setHeader("Content-Disposition", 
						"attachment;filename="+HTML.urlEncode(title)+"." + format);
				}
				response.setContentType("image/" + format);
				response.setHeader("Cache-Control", "max-age=3600");
				OutputStream out = response.getOutputStream();
				ImageIO.write(image, format, out);
				out.flush();
				out.close();
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
		

		
	}

	private void saveImageToFile(DbControl dc, BufferedImage image, String path, 
		String format, boolean overwrite)
		throws IOException
	{
		Path p = new Path(path, Path.Type.FILE);
		File f = File.getByPath(dc, p, true);
		f.setMimeType("image/"+format);
		if (f.isInDatabase())
		{
			if (!overwrite) throw new ItemAlreadyExistsException("File[path="+path+"]");
		}
		else
		{
			dc.saveItem(f);
		}
		OutputStream out = f.getUploadStream(false);
		if (image != null)
		{
			ImageIO.write(image, format, out);
		}
		out.flush();
		out.close();
	}
	
	/**
		Creates a box-and-whisker data holder for the given list of values.
		The data holder will be populated with the median and the 5th, 25th, 75th and
		95th percentile values as follows:
		<ul>
		<li>median = BAWI.getMedian()
		<li>5th percentile = BAWI.getMinRegularValue()
		<li>25th percentile = BAWI.getQ1()
		<li>75th percentile = BAWI.getQ3()
		<li>95th percentile = BAWI.getMaxRegularValue()
		</ul>
		The mean and outlier values are currently not calculated (may change in
		the future).
		
		@param values The list of values which must be sorted in ascending order
		@return A BoxAndWhiskerItem object
	*/
	public BoxAndWhiskerItem calculateBoxAndWhiskerStatistics(List<Float> values)
	{
		float median = getPercentile(values, 50);
		float p5 = getPercentile(values, 5);
		float p25 = getPercentile(values, 25);
		float p75 = getPercentile(values, 75);
		float p95 = getPercentile(values, 95);
		return new BoxAndWhiskerItem(Float.NaN, median, p25, p75, p5, p95, Float.NaN, Float.NaN, null);
	}
	
	/**
		Calculates the percentile values from a sorted list of values.
		If the percentile falls between two elements in the list the
		average of the two nearest elements are used.
		
		@param values The values which must be sorted
		@param percentile The percentile, a value between 0 and 100
		@return The percentile value, or Float.NaN if it could not be
			calculated
	*/
	public float getPercentile(List<? extends Number> values, int percentile)
	{
		int size = values.size();
		if (size == 0) return Float.NaN;
		if (size == 1) return values.get(0).floatValue();
		
		int width = size + 1;
		int pos = width * percentile / 100;
		if (pos == 0)
		{
			 // Overflow to the left -- return first value
			return values.get(0).floatValue();
		}
		else if (pos >= size)
		{
			 // Overflow to the right -- return last value
			return values.get(size-1).floatValue();
		}
		
		if (width * percentile % 100 == 0)
		{
			// Exact hit -- return one value
			return values.get(pos-1).floatValue();
		}
		else
		{
			// Hit between two values -- return the average
			return (values.get(pos-1).floatValue() + values.get(pos).floatValue())/2;
		}
	}
	
	/**
		Uses as a column key in the plots. This allows us to
		separate the displayed label from the identity of the column.
		Eg. there is no need to make sure that the label strings are unique
		since we can use some internal id for this purpose.
		<p>
		Additionally, this class allows us to map annotations to colors
		that are used to draw small colored boxes below each column label.
	*/
	public static class ColumnKey
		implements Comparable<ColumnKey>
	{
		private final Comparable id;
		private final String title;
		private final Map<String, Color> annotations;
		
		public ColumnKey(Comparable id, String title)
		{
			this.id = id;
			this.title = title;
			this.annotations = new HashMap<String, Color>();
		}

		/*
			From the Comparable interface
			-----------------------------
		*/
		/**
			The ID of the key determines equality.
		*/
		@SuppressWarnings("unchecked")
		@Override
		public int compareTo(ColumnKey o)
		{
			return id.compareTo(o.id);
		}
		// -------------------------------
		
		/*
			From the Object class
			-------------------------------
		*/
		@Override
		public String toString()
		{
			return title;
		}
		
		@Override
		public boolean equals(Object o)
		{
			if (o == null) return false;
			if (this == o) return true;
			if (o.getClass() != getClass()) return false;
			ColumnKey ck = (ColumnKey)o;
			return id.equals(ck.id);
		}
		
		@Override
		public int hashCode()
		{
			return id.hashCode();
		}
		// ------------------------------------
		
		public void setColor(String annotation, Color color)
		{
			annotations.put(annotation, color);
		}
		
		public Color getColor(String annotation)
		{
			return annotations.get(annotation);
		}
		
	}
	
}
