/*
	$Id: Note.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;

/**
	This tag is used to display small notes on a BASE web page.
	<p>
	<b>Example:</b>
	<pre>
   &lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
   &lt;base:note type="info" title="Some information"&gt;
   Here is some important information...
   &lt;/base:note&gt;
</pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:note
      id=...
      clazz=...
      style=...
      messagestyle=...
      type=info|information|warning|question|error
      title=...
      icon=...
      visible=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>The ID of the note. The value of this attribute
		goes directly into the standard <code>id</code> attribute.
		This means that if you need a reference to the note object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		note = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>note</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different notes.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different notes.
		</td>
	</tr>
	<tr>
		<td>messagestyle</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute of the message part. This allows you to apply different styles
			to different notes.
		</td>
	</tr>
	<tr>
		<td>type</td>
		<td>error</td>
		<td>yes</td>
		<td>The type of note to display. Four types are pre-defined:
			<ul>
			<li>info,information: An information message.
			<li>warning: A warning message.
			<li>question: A question.
			<li>error: An error message.
			</ul>
			The type is used to select a propert <code>icon</code>
			unless it is specified.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td></td>
		<td>no</td>
		<td>The title of the note.</td>
	</tr>
	<tr>
		<td>icon</td>
		<td></td>
		<td>no</td>
		<td>
			The filename of an icon to display together with the message. If
			not specified a default icon matching the type attribute is 
			automatically selected. The icon file must be located
			in the <code>/images</code> directory.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the note should be visible or not.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Note
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = 7490498743887372136L;

	private static final int NOTE_TYPE_ERROR = 1;
	
	private static final int NOTE_TYPE_INFORMATION = 2;

	private static final int NOTE_TYPE_WARNING = 3;
	
	private static final int NOTE_TYPE_QUESTION = 4;

	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;

	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = "note";
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;

	/**
		Optional <code>style</code> attribute for the message section
	*/
	private String messageStyle = null;

	
	/**
		The type of note.
	*/
	private int type = NOTE_TYPE_ERROR;

	/**
		The title of the note.
	*/
	private String title = null;

	/**
		The filename of an optional icon that should be displayed with the error message.
	*/
	private String icon = null;

	/**
		If the note should be visible or not.
	*/
	private boolean visible = true;
	
	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}	

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}
	
	public void setMessagestyle(String messageStyle)
	{
		this.messageStyle = messageStyle;
	}
	public String getMessagestyle()
	{
		return messageStyle;
	}

	public void setType(String type)
	{
		if ("info".equalsIgnoreCase(type) || "information".equalsIgnoreCase(type))
		{
			this.type = NOTE_TYPE_INFORMATION;
		}
		else if ("warning".equalsIgnoreCase(type))
		{
			this.type = NOTE_TYPE_WARNING;
		}
		else if ("question".equalsIgnoreCase(type))
		{
			this.type = NOTE_TYPE_QUESTION;
		}
		else
		{
			this.type = NOTE_TYPE_ERROR;
		}
	}

	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}

	public void setIcon(String icon)
	{
		this.icon = icon;
	}
	public String getIcon()
	{
		return icon;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}

	public int doStartTag()
		throws JspException
	{
		page = (Page)findAncestorWithClass(this, Page.class);
		if (page == null) throw new JspTagException("Tag <base:note> must be inside a <base:page> tag");
		if (!isVisible()) return SKIP_BODY;

		String theIcon = getIcon();
		if (theIcon == null)
		{
			switch (type)
			{
			case NOTE_TYPE_ERROR:
				theIcon = "error.gif";
				break;
			case NOTE_TYPE_WARNING:
				theIcon = "warning.gif";
				break;
			case NOTE_TYPE_QUESTION:
				theIcon = "question.gif";
				break;
			default:
				theIcon = "info.gif";
				break;
			}
		}

		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		if (getId() != null) sb.append(" id=\"").append(getId()).append("\"");
		if (getClazz() != null) sb.append(" class=\"").append(getClazz()).append("\"");
		if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
		sb.append(">\n");
		if (getTitle() != null) sb.append("<h3 class=\"docked\">").append(getTitle()).append("</h3>\n");
		sb.append("<table class=\"main\" cellspacing=0 cellpadding=0 width=\"100%\">\n");
		sb.append("<tr>\n");
		if (theIcon != null)
		{
			sb.append("<td class=\"icon\" rowspan=2>");
			sb.append("<img src=\"").append(page.getRoot()).append("images/").append(theIcon).append("\"></td>\n");
		}
		sb.append("<td class=\"message\"><div");
		if (getMessagestyle() != null) sb.append(" style=\"").append(getMessagestyle()).append("\"");
		sb.append(">");

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	public int doEndTag()
		throws JspException
	{
		if (!isVisible()) return EVAL_PAGE;
		StringBuilder sb = new StringBuilder();
		sb.append("</div></td>\n\t</tr>\n</table>\n</div>\n");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}
}




