/*
	$Id: Cell.java 5518 2010-11-22 08:02:13Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.tagext.BodyTagSupport;
import javax.servlet.jsp.tagext.TryCatchFinally;

import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.util.formatter.Formatter;

/**
	This tag defines a new cell in a row. It should be placed inside a
	{@link Row &lt;tbl:row&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:cell
      clazz=...
      errorclazz=...
      style=...
      column=...
      visible=true|false
      value=...
      maxcharacters=...
      maxheight=...
      maxwidth=...
      autoshow=true|false
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>cell</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>errorclazz</td>
		<td>cell error</td>
		<td>no</td>
		<td>
			Value for the class attribute to use if an exception is thrown
			while evaluating the contents of the cell. Since BASE 2.9.1
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>column</td>
		<td>-</td>
		<td>yes</td>
		<td>
			The ID of the {@link ColumnDef &lt;tbl:columndef&gt;} this cell
			is linked to.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the cell should be visible or not.
		</td>
	</tr>
	<tr>
		<td>value</td>
		<td>-</td>
		<td>no</td>
		<td>
			A value to display in the cell. If a {@link Formatter} has been
			defined for this column the value is formatted otherwise it is just
			converted to a string with the toString() method.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Cell
	extends BodyTagSupport
	implements TryCatchFinally
{

	/**
		
	*/
	private static final long serialVersionUID = -1992298892993613884L;

	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;
	
	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = "cell";

	/**
		Optional <code>class</code> attribute for errors.
	*/
	private String errorClazz = "cell error";

	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;

	/**
		The ID of the column to which this cell belongs.
	*/
	private String columnId;

	/**
		If the cell should be visible or not.
	*/
	private boolean visible = true;
	
	/**
		The value to display.
	*/
	private transient Object value = null;

	private transient Formatter formatter;
	
	private boolean isHidden;
	
	private int maxCharacters = 50;
	private boolean autoShow = true;
	private String maxHeight = null;
	private String maxWidth = null;
	
	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}
	
	/**
		@since 2.9.1
	*/
	public void setErrorclazz(String errorClazz)
	{
		this.errorClazz = errorClazz;
	}
	/**
		@since 2.9.1
	*/
	public String getErrorclazz()
	{
		return errorClazz;
	}

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}
	
	public void setColumn(String columnId)
	{
		this.columnId = columnId;
	}
	public String getColumn()
	{
		return columnId;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}
	
	public void setValue(Object value)
	{
		this.value = value;
	}
	
	public Object getValue()
	{
		return value;
	}

	public Formatter<?> getFormatter()
	{
		return formatter;
	}
	
	public void setMaxcharacters(int maxCharacters)
	{
		this.maxCharacters = maxCharacters;
	}
	
	public void setAutoshow(boolean autoShow)
	{
		this.autoShow = autoShow;
	}
	
	public void setMaxwidth(String maxWidth)
	{
		this.maxWidth = maxWidth;
	}
	
	public void setMaxheight(String maxHeight)
	{
		this.maxHeight = maxHeight;
	}
	
	/*
		Display methods
	*/
	public int doStartTag()
		throws JspException
	{
		table = (Table)findAncestorWithClass(this, Table.class);
		isHidden = !isVisible() || !table.isColumnVisible(columnId);
		if (isHidden) return SKIP_BODY;
		formatter = table.getFormatter(columnId);
		return EVAL_BODY_BUFFERED;
	}

	@SuppressWarnings("unchecked")
	public int doEndTag()
		throws JspException
	{
		if (isHidden) return EVAL_PAGE;
		String content = bodyContent == null ? "" : bodyContent.getString();
		content = content == null ? "" : content.trim();
		if (value != null)
		{
			content += formatter == null ? value.toString() : formatter.format(value);
		}
		if (content.length() == 0) content = "&nbsp;";
		boolean overflowed = maxCharacters > 0 && content.length() > maxCharacters && 
			HTML.textLength(content) > maxCharacters;

		StringBuilder sb = new StringBuilder();
		sb.append("\t<td");
		if (getClazz() != null) sb.append(" class=\"").append(getClazz()).append("\"");
		if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
		if (getId() != null) sb.append(" id=\"").append(getId()).append("\"");
		sb.append(">");
		if (overflowed)
		{
			sb.append("<div class=\"constrained unchecked");
			if (autoShow) sb.append(" autoshow");
			sb.append("\"");
			if (maxWidth != null || maxHeight != null)
			{
				sb.append(" style=\"");
				if (maxWidth != null) sb.append("max-width: ").append(maxWidth).append(";");
				if (maxHeight != null) sb.append("max-height: ").append(maxHeight).append(";");
				sb.append("\"");
			}
			sb.append(">");
		}
		sb.append(content);
		if (overflowed) sb.append("</div>");
		sb.append("</td>\n");

		table.setColumnContent(columnId, sb.toString());
		return EVAL_PAGE;
	}

	/*
		From the TryCatchFinally interface
		----------------------------------
	 */
	@Override
	public void doCatch(Throwable t) 
		throws Throwable
	{
		t.printStackTrace();
		StringBuilder sb = new StringBuilder();
		sb.append("\t<td");
		if (getErrorclazz() != null) sb.append(" class=\"").append(getErrorclazz()).append("\"");
		if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
		sb.append(">");
		sb.append("<span title=\"See log file for more information about this error\">");
		String msg = t.getMessage();
		if (msg == null) msg = t.getClass().getName();
		sb.append(msg).append("</span>");
		sb.append("</td>\n");
		table.setColumnContent(columnId, sb.toString());
	}	
	
	@Override
	public void doFinally()
	{}
	// ------------------------------------

	
}


