/*
	$Id: AbstractBatcher.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.lang.ref.WeakReference;
import java.sql.SQLException;

import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;

/**
	This is an abstract class for batchers that are connected to a
	single {@link DbControl} object. It implements all methods in the
	{@link Batcher} interface except the {@link Batcher#flush()} method.
	<p>
	A subclass may also have to override the {@link #close()} method
	if it has opened connections or statements to the database.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public abstract class AbstractBatcher
	implements Batcher
{

	/**
		Log all SQL statements.
	*/
	protected static final org.apache.log4j.Logger logSql = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.batcher.sql");
	
	/**
		So we don't always have to call logSql.debug()
	*/
	protected static final boolean debugSqlEnabled = logSql.isDebugEnabled();

	
	private SessionControl sc;

	/**
		A reference to the current DbControl object. We use a WeakReference since
		we it is always the client applications responsibility to keep control
		over the DbControl object.
	*/
	private WeakReference<DbControl> dc;

	/**
		SQL command batch size.
	*/
	private int batchSize;
	
	/**
		Constructor for a batcher.
		@throws BaseException If there is an error
	*/
	AbstractBatcher()
	{
		this.batchSize = Config.getInt("db.batch-size", 50);
	}

	/*
		From the Batcher interface
		-------------------------------------------
	*/
	public int getBatchSize()
	{
		return batchSize;
	}
	public void setBatchSize(int batchSize)
	{
		this.batchSize = batchSize;
	}
	/**
		Check if this batcher is closed or not. The batcher is considered as closed
		when it isn't attached to a {@link DbControl} or if the <code>DbControl</code>
		it is attached to has been closed.

		@return TRUE if the batcher is closed, FALSE otherwise
	*/
	public final boolean isClosed()
	{
		DbControl tmp = dc == null ? null : dc.get();
		return tmp == null || tmp.isClosed();
	}
	public final void close()
		throws BaseException
	{
		DbControl tmp = dc == null ? null : dc.get();
		if (tmp != null && !tmp.isClosed())
		{
			if (!tmp.isRolledBack()) flush();
			onBeforeClose();
			tmp.removeBatcher(this);
			dc = null;
			sc = null;
		}
	}
	// -------------------------------------------

	/**
		Update the last access time so the session control doesn't 
		get cleaned up.
	*/
	void updateLastAccess()
	{
		sc.updateLastAccess();
	}
	
	/**
		Set the DbControl that is controlling this batcher. This
		method MUST be called by all subclasses after successful construction
		of a batcher. This method also registers the batcher with the DbControl
		using {@link DbControl#addBatcher(Batcher)}
		@since 2.4.2
	*/
	protected final void setDbControl(DbControl dc)
	{
		assert dc != null : "dc == null";
		this.dc = new WeakReference<DbControl>(dc);
		this.sc = dc.getSessionControl();
		dc.addBatcher(this);
	}
	
	/**
		Get the {@link DbControl} object that currently manages this batcher.
		@throws ConnectionClosedException If the batcher is connected to a closed 
			<code>DbControl</code> object, or not connected at all
	*/
	public final DbControl getDbControl()
		throws ConnectionClosedException
	{
		DbControl tmp = dc == null ? null : dc.get();
		if (tmp == null || tmp.isClosed()) throw new ConnectionClosedException();
		return tmp;
	}

	/**
		Get the {@link SessionControl} object that manages this item.
		@throws ConnectionClosedException If the item is connected to a closed 
			<code>DbControl</code> object, or not connected at all
	*/
	public final SessionControl getSessionControl()
		throws ConnectionClosedException
	{
		return sc;
	}
	
	/**
		Let a subclass cleanup itself before the connection to
		the database is closed. This method is called by the {@link #close()}
		method after {@link #flush()} has been called but before the connection
		to the database is lost. Typical use of this method is to update 
		properties on parent object (for example, spot count) and to close
		open SQL statements.
	*/
	void onBeforeClose()
		throws BaseException
	{}
	
	/**
		Update index statistics for a table. 
		@see DbEngine#analyzeAfterBatchOperation()
	 */
	void analyzeTable(String catalog, String schema, String table)
	{
		DbEngine dbe = HibernateUtil.getDbEngine();
		if (dbe.analyzeAfterBatchOperation())
		{
			String sql = dbe.getAnalyzeTableSql(catalog, schema, table);
			if (sql != null)
			{
				if (debugSqlEnabled) 
				{
					logSql.debug("Analyzing table " + catalog + "." + schema + "." + table +": " + sql);
				}
				try
				{
					HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(), new ExecuteUpdateWork(sc, sql));
				}
				catch (SQLException ex)
				{
					logSql.error(sql, ex);
					LogUtil.logSQLExceptionChain(logSql, ex);
					throw new BaseException(ex);
				}
			}
		}
	}

}

