/*
	$Id: Include.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.Collections;
import java.util.Set;
import java.util.EnumSet;

/**
	This enumeration defines constants that are used to specify which items
	to include in a query. 

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.EntityQuery#include(Include[])
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public enum Include
{
	/**
		Include items flagged as removed.
	*/
	REMOVED(1),
	/**
		Include items flagged as not removed.
	*/
	NOT_REMOVED(2),
	/**
		Include items that have at least one annotation.
	*/
	ANNOTATED(4),
	/**
		Include items that have no annotations.
	*/
	NOT_ANNOTATED(8),
	/**
		Include item owned by the logged in user.
	*/
	MINE(16),
	/**
		Include items owned by other users. Requires generic read permission for the type of item.
	*/
	OTHERS(32),
	/**
		Include items shared to the logged in user.
	*/
	SHARED(64),
	/**
		Include items in the active project.
	*/
	IN_PROJECT(128);

	/**
		A non-modifable collection containing all options.
		@since 2.5
	*/
	public static final Collection<Include> ALL = Collections.unmodifiableSet(EnumSet.allOf(Include.class));
	
	private final int value;
	
	/**
		Constructor for Permission enums.
	*/
	private Include(int value)
	{
		this.value = value;
	}

	private int getValue()
	{
		return value;
	}

	
	public static int toInt(Set<Include> includes)
	{
		int result = 0;
		if (includes != null)
		{
			for (Include i : includes)
			{
				result |= i.getValue();
			}
		}
		return result;
	}
	
	/**
		Convert an integer value to a set of includes.
	*/
	static Set<Include> fromInt(int includes)
	{
		EnumSet<Include> result = EnumSet.noneOf(Include.class);
		for (Include i : Include.values())
		{
			if ((includes & i.getValue()) == i.getValue())
			{
				result.add(i);
			}
		}
		return result;
	}

}
