/*
	$Id: ReporterPropertyExpression.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.util.EqualsHelper;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

/**
	An expression representing the property of a
	reporter.
	<p>
	Throws an <code>UnsupportedOperationException</code> if 
	{@link Query#getQueryType()} isn't {@link QueryType#SQL}.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Dynamic#reporter(String)
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class ReporterPropertyExpression
	implements Expression
{
	private static volatile Map<String, String> propertyToColumn;

	private final String property;
	
	private final String column;

	public ReporterPropertyExpression(String property)
		throws InvalidDataException
	{
		this.property = property;
		this.column = getColumn(property);
		if (column == null) throw new InvalidDataException("Unknown property '"+property+"' of reporter.");
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			return HibernateUtil.quote(RealTable.REPORTERS.getAlias()) + "." + HibernateUtil.quote(column); 
		}
		else
		{
			throw new UnsupportedOperationException("'reporterProperty' operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return RealTable.REPORTERS.getAlias() + "." + property; 
	}
	/**
		This expression is equal to another ReporterPropertyExpression if they 
		represent the same reporter property.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		ReporterPropertyExpression o = (ReporterPropertyExpression)other;
		return EqualsHelper.equals(property, o.property);
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(property);
	}
	// -------------------------------------------
	
	private static String getColumn(String property)
	{
		if (propertyToColumn == null)
		{
			synchronized (ReporterPropertyExpression.class)
			{
				// So we are really sure no other thread created this while we were waiting
				if (propertyToColumn == null)
				{
					propertyToColumn = new HashMap<String, String>();
					org.hibernate.mapping.PersistentClass pc = HibernateUtil.getClassMapping("net.sf.basedb.core.data.ReporterData");
					Iterator propIt = pc.getPropertyIterator();
					while (propIt.hasNext())
					{
						org.hibernate.mapping.Property prop = (org.hibernate.mapping.Property)propIt.next();
						Iterator colIt = prop.getColumnIterator();
						if (colIt.hasNext())
						{
							org.hibernate.mapping.Column col = (org.hibernate.mapping.Column)colIt.next();
							propertyToColumn.put(prop.getName(), col.getName());
						}
					}
					propertyToColumn.put("id", "id"); //Add id as a property because hibernate wont
				}
			}
		}
		return propertyToColumn.get(property);
	}
}
