/*
  $Id $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.QuantityData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.data.UnitSymbolData;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.PrefixSuffixFormatter;

/**
	This class is used to represent a specific unit that can be
	used with annotations. A unit must be related to the
	reference unit of the quantity. See {@link Quantity#getReferenceUnit()}
	for examples.

	@author Nicklas
	@version 2.9
	@base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
	@see Quantity
	@see <a href="http://physics.nist.gov/cuu/Units/units.html">Essentials of 
		the SI: Base &amp; derived units</a>
*/
public class Unit
	extends BasicItem<UnitData>
	implements Nameable, Removable
{

	/**
		The type of item represented by this class.
		@see Item#UNIT
		@see #getType()
	*/
	public static final Item TYPE = Item.UNIT;
	
	/**
		Create a new <code>Unit</code> item. The unit will have 
		default values for the reference factor (1.0) and offset (0.0).
		Use {@link #changeReferenceFactorAndOffset(double, double)}
		to change those.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param quantity The quantity the unit belongs to
		@param displaySymbol The symbol used when displaying values with 
			this unit
		@return The new <code>Unit</code> item
		@throws BaseException If there is an error
	*/
	public static Unit getNew(DbControl dc, Quantity quantity, String displaySymbol)
		throws BaseException
	{
		Unit u = dc.newItem(Unit.class);
		u.setName("New unit");
		u.setQuantity(quantity);
		u.setDisplaySymbol(displaySymbol);
		u.getData().setReferenceFactor(1);
		u.getData().setReferenceOffset(0);
		return u;
	}
	
	/**
		Get a <code>Unit</code> object when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The system ID of the item to load.
		@return The <code>Unit</code> item
		@throws ItemNotFoundException This exception is thrown if an
			item with the specified ID is not found
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have read
			permission to the item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static Unit getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Unit u = dc.loadItem(Unit.class, id);
		if (u == null) throw new ItemNotFoundException("Unit[ID="+id+"]");
		return u;
	}
	
	/**
		Get a query configured to retrieve units.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Unit> getQuery()
	{
		return new ItemQuery<Unit>(Unit.class);
	}
	
	/**
		Creates a unit item.
		@param data the data
	*/
	Unit(UnitData data)
	{
		super(data);
	}

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>Annotation types are using this unit as default unit
		<li>Annotations are using this unit
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ANNOTATIONS_FOR_UNIT", "count(*)");
		/*
			SELECT {1}
			FROM AnnotationData a
			WHERE a.unit = :unit
		*/
		query.setEntity("unit", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ANNOTATIONTYPES_FOR_UNIT", "count(*)");
			/*
				SELECT {1}
				FROM AnnotationTypeData at
				WHERE at.defaultUnit = :unit
			*/
			query.setEntity("unit", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link AnnotationType}:s using this unit as default unit
		<li>{@link Annotation}:s using this unit
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Annotation types
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ANNOTATIONTYPES_FOR_UNIT", "at.id");
		/*
				SELECT {1}
				FROM AnnotationTypeData at
				WHERE at.defaultUnit = :unit
		*/
		query.setEntity("unit", this.getData());
		addUsingItems(using, Item.ANNOTATIONTYPE, query);
		
		// Annotations
		query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ANNOTATIONS_FOR_UNIT", "a.id");
		/*
				SELECT {1}
				FROM AnnotationData a
				WHERE a.unit = :unit
		*/
		query.setEntity("unit", this.getData());
		addUsingItems(using, Item.ANNOTATION, query);
		return using;
	}
	
	/**
		Set property filters that use this unit to null
		when the unit is deleted.
	*/
	@Override
	void onBeforeCommit(Action action) 
		throws BaseException 
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.DELETE)
		{
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(
				getDbControl().getHibernateSession(), "SET_UNIT_TO_NULL_ON_PROPERTY_FILTERS");
			/*
				UPDATE [PropertyFilters]
				SET [unit_id] = null 
				WHERE [unit_id] = :unit
			*/
			query.setInteger("unit", getId());
			HibernateUtil.executeUpdate(query);
		}
	}

	// -------------------------------------------
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	//  -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}

	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}

	public String getDescription()
	{
		return getData().getDescription();
	}

	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------

	/**
		Get the quantity this unit belongs to.
	*/
	public Quantity getQuantity()
	{
		return getDbControl().getItem(Quantity.class, getData().getQuantity());
	}
	
	private void setQuantity(Quantity quantity)
	{
		if (quantity == null) throw new InvalidUseOfNullException("quantity");
		getData().setQuantity(quantity.getData());
	}
	
	/**
		The maximum length of the display symbol that can be stored in the database.
		@see #setDisplaySymbol(String)
	*/
	public static final int MAX_SYMBOL_LENGTH = UnitData.MAX_SYMBOL_LENGTH;
	/**
		Get the display symbol to use when displaying values with this
		unit.
	 	@return A String object 
	*/
	public String getDisplaySymbol()
	{
		return getData().getDisplaySymbol();
	}
	
	/**
		Set the display symbol for this unit. The symbol is also registered
		as an alias symbol for this unit. The old symbol is NOT removed as
		an alias. Use {@link #removeSymbol(String)} to do that.
		
	 	@param symbol The symbol
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value null or longer than
			{@link #MAX_SYMBOL_LENGTH}
	*/
	public void setDisplaySymbol(String symbol)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		String oldSymbol = getData().getDisplaySymbol();
		getData().setDisplaySymbol(StringUtil.setNotNullString(symbol, "symbol", MAX_SYMBOL_LENGTH));
		if (!symbol.equals(oldSymbol)) addSymbol(symbol);
	}

	/**
		Get all symbols that are registered for this unit.
		@return All symbols as a set
	*/
	public Set<String> getSymbols()
	{
		Set<String> symbols = new TreeSet<String>();
		for (UnitSymbolData s : getData().getSymbols())
		{
			symbols.add(s.getSymbol());
		}
		return symbols;
	}
	
	/**
		Set the symbols that should be registered for this unit. Note that
		all units must be unique within a given {@link Quantity}. Symbols that
		are not in the given set will be unregistered except for the display
		symbol.
		@param symbols A set with the symbols
	*/
	public void setSymbols(Set<String> symbols)
	{
		checkPermission(Permission.WRITE);
		if (symbols == null) throw new InvalidUseOfNullException("symbols");
		Set<String> newSymbols = new HashSet<String>(symbols);
		// The display symbold must always be present
		newSymbols.add(getData().getDisplaySymbol());
		Iterator<UnitSymbolData> currentSymbols = getData().getSymbols().iterator();
		while (currentSymbols.hasNext())
		{
			UnitSymbolData c = currentSymbols.next();
			if (newSymbols.contains(c.getSymbol()))
			{
				// This symbol is already registered, do nothing
				newSymbols.remove(c.getSymbol());
			}
			else
			{
				// This symbol is registered, but is not present among the
				// new symbols --> remove it
				currentSymbols.remove();
			}
		}
		UnitData data = getData();
		QuantityData quantity = data.getQuantity();
		for (String newSymbol : newSymbols)
		{
			UnitSymbolData s = new UnitSymbolData();
			s.setSymbol(newSymbol);
			s.setUnit(data);
			s.setQuantity(quantity);
			data.getSymbols().add(s);
		}
	}
	
	/**
		Add a symbol to this unit.
		@param symbol The symbol to add
	*/
	public void addSymbol(String symbol)
	{
		checkPermission(Permission.WRITE);
		if (symbol == null) throw new InvalidUseOfNullException("symbol");
		Set<UnitSymbolData> symbols = getData().getSymbols();
		for (UnitSymbolData s : symbols)
		{
			if (s.getSymbol().equals(symbol)) return;
		}
		UnitSymbolData s = new UnitSymbolData();
		s.setSymbol(symbol);
		s.setUnit(getData());
		s.setQuantity(getData().getQuantity());
		symbols.add(s);
	}
	
	/**
		Remove a symbol from this unit. It is not possible to remove
		the display symbol.
		@param symbol The symbol to remove
	*/
	public void removeSymbol(String symbol)
	{
		checkPermission(Permission.WRITE);
		if (symbol == null) throw new InvalidUseOfNullException("symbol");
		if (symbol.equals(getDisplaySymbol())) return;
		Iterator<UnitSymbolData> symbols = getData().getSymbols().iterator();
		while (symbols.hasNext())
		{
			UnitSymbolData s = symbols.next();
			if (s.getSymbol().equals(symbol)) 
			{
				symbols.remove();
				return;
			}
		}
	}
	
	/**
		Get the offset that is used for converting values in this
		unit into the reference unit of the quantity. See
		{@link Quantity#getReferenceUnit()} for a description of how the
		values are converted.
		@return The reference offset
		@see #changeReferenceFactorAndOffset(double, double)
	*/
	public double getReferenceOffset()
	{
		return getData().getReferenceOffset();
	}
	
	/**
		Get the factor that is used for converting values in this
		unit into the reference unit of the quantity. See
		{@link Quantity#getReferenceUnit()} for a description of how the
		values are converted.
		@return The reference factor
		@see #changeReferenceFactorAndOffset(double, double)
	*/
	public double getReferenceFactor()
	{
		return getData().getReferenceFactor();
	}
	
	/**
		Changes the factor and offset used for converting values in this
		unit into the reference unit. Calling this method with values that
		are different from the old ones will result in a re-calculation of all 
		annotation values that has used this unit to make sure that the actual 
		values stays the same.
		@param factor The new factor
		@param offset The new offset
	*/
	public void changeReferenceFactorAndOffset(double factor, double offset)
	{
		checkPermission(Permission.WRITE);
		if (Double.isNaN(factor)) 
		{
			throw new NumberOutOfRangeException("NaN is not allowed for 'factor'");
		}
		if (Double.isNaN(offset))
		{
			throw new NumberOutOfRangeException("NaN is not allowed for 'offset'");
		}
		UnitData data = getData();
		if (isInDatabase())
		{
			boolean sameFactor = Double.compare(data.getReferenceFactor(), factor) == 0;
			boolean sameOffset = Double.compare(data.getReferenceOffset(), offset) == 0;
			if (!sameFactor || !sameOffset)
			{
				/*
					Need to recalculate annotation values for existing annotations 
					using this unit. There are two cases. Both are re-calculated 
					with newValue = aFactor * oldValue + anOffset:
					1. Annotations with this unit with an annotation type of a different
					   default unit.
					   aFactor = newFactor / oldFactor
					   anOffset = newOffset - oldOffset * aFactor
					2. Annotation with a different unit that has an annotation type with
					   this unit as default unit.
					   aFactor = oldFactor / newFactor
					   anOffset = (oldOffset - newOffset) / newFactor
					We need to execute all SQL:s on all numerical value
					tables.
				*/
				double case1Factor = factor / getReferenceFactor();
				double case2Factor = getReferenceFactor() / factor;
				double case1Offset = offset - getReferenceOffset() * case1Factor;
				double case2Offset = (getReferenceOffset() - offset) / factor;
				
				for (Type valueType : Type.values())
				{
					if (!valueType.isNumerical())
					{
						/* #### CONTINUE-STATEMENT #### */
						continue;
					}
					
					String tableName = valueType.getTableName();
					org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(
						getDbControl().getHibernateSession(), 
						"RECALCULATE_ANNOTATIONS_FOR_UNIT", tableName);
					/*
						UPDATE [{1}] SET [value] = [value] * :factor + :offset
						WHERE [id] IN 
						(
							SELECT [a].[value_id]
							FROM [Annotations] [a]
							JOIN [AnnotationTypes] [at] ON [at].[id] = [a].[annotationtype_id]
							WHERE [at].[value_type] = :valueType
							AND 
							(
								(:case = 1 AND [a].[unit_id] = :unit AND [at].[default_unit_id] <> :unit)
								OR
								(:case = 2 AND [a].[unit_id] <> :unit AND [at].[default_unit_id] = :unit)
							)
						)
					*/
					query.setInteger("unit", getId());
					query.setInteger("valueType", valueType.getValue());
					
					// Case 1
					query.setInteger("case", 1);
					query.setDouble("factor", case1Factor);
					query.setDouble("offset", case1Offset);
					HibernateUtil.executeUpdate(query);
					
					// Case 2
					query.setInteger("case", 2);
					query.setDouble("factor", case2Factor);
					query.setDouble("offset", case2Offset);
					HibernateUtil.executeUpdate(query);
				}
			}
		}
		data.setReferenceFactor(factor);
		data.setReferenceOffset(offset);
	}

	/**
		Convert a value from this unit to the reference unit.
		@param value The value in this unit
		@return The value converted to the reference unit
	*/
	public double toReferenceUnit(double value)
	{
		return value * getData().getReferenceFactor() + getData().getReferenceOffset();
	}
	
	/**
		Convert a value from the reference unit to this unit.
		@param value The reference value
		@return The value converted to this unit
	*/
	public double fromReferenceUnit(double value)
	{
		return (value - getData().getReferenceOffset()) / getData().getReferenceFactor();
	}
	
	/**
		Convert a value from this unit to another unit. This method is
		equivalent to calling 
		<code>other.fromReferenceUnit(this.toReferenceUnit(value))</code>.
		
		@param value The value in this unit
		@param other The other unit
		@return The value converted to the other unit
	*/
	public double toUnit(double value, Unit other)
	{
		return other.fromReferenceUnit(this.toReferenceUnit(value));
	}
	
	/**
		Convert a value from another unit to this unit. This method is equivalent
		to calling <code>this.fromReferenceUnit(other.toReferenceUnit(value))</code>
		
		@param value The value in the other unit
		@param other The other unit
		@return The value converted to this unit
	*/
	public double fromOther(double value, Unit other)
	{
		return this.fromReferenceUnit(other.toReferenceUnit(value));
	}
	
	/**
		Create a formatter for formatting values with this unit. The returned
		formatter delegates the formatting of the values to a parent formatter
		and uses a {@link PrefixSuffixFormatter} to append the unit symbol.
		
		@param parent The parent formatter for formatting the actual values
		@return The new formatter
	*/
	public <T> Formatter<T> getFormatter(Formatter<T> parent)
	{
		return new PrefixSuffixFormatter<T>("", parent, 
				PrefixSuffixFormatter.NBSP + getDisplaySymbol());
	}
	
	/**
		Find the unit with a given symbol that has the same
		quantity as this unit.
		@return The unit, or null if no unit is found
	*/
	UnitData findOtherUnit(org.hibernate.Session session, String symbol)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(
				session, "GET_UNIT_WITH_SYMBOL_FOR_UNIT");
		query.setEntity("unit", this.getData());
		query.setString("symbol", symbol);
		UnitData unit = HibernateUtil.loadData(UnitData.class, query);
		return unit;
	}

}
