/*
	$Id: Update.java 5858 2011-11-09 14:01:29Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Table;


import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnyToAnyData;
import net.sf.basedb.core.data.ArraySlideData;
import net.sf.basedb.core.data.BioMaterialEventSourceData;
import net.sf.basedb.core.data.ContextData;
import net.sf.basedb.core.data.DataFileTypeData;
import net.sf.basedb.core.data.DerivedBioAssayData;
import net.sf.basedb.core.data.FileData;
import net.sf.basedb.core.data.FileSetData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.data.HardwareData;
import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ItemParameterValueData;
import net.sf.basedb.core.data.ItemSubtypeData;
import net.sf.basedb.core.data.PhysicalBioAssayData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.ProtocolData;
import net.sf.basedb.core.data.RoleKeyData;
import net.sf.basedb.core.data.SchemaVersionData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.hibernate.JdbcWork;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bcrypt.BCrypt;

/**
	This class contains static methods used to update items already in the database.
	
	<table>
	<tr>
		<th>Schema version</th>
		<th>Description</th>
	</tr>
	
	<tr>
		<td>90</td>
		<td>
			The schema as released in BASE 2.17. For information about schema changes before
			this check the BASE 2.17 code.
		</td>
	</tr>
	
	<tr>
		<td>100</td>
		<td>
			The schema as released in BASE 3.0. We don't list all changes here.
		</td>
	</tr>
	<tr>
		<td>101 and 102</td>
		<td>
			Fixes an issue with duplicate {@link FileSetMemberData} entries
			due to a bug in the upgrade program. Make sure that the unique
			constraint that should prevent duplicates exists.
		</td>
	</tr>

	</table>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-11-09 15:01:29 +0100 (Wed, 09 Nov 2011) $
*/
public final class Update
{
	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.Update");

	/**
		Update the database by modifying existing items to follow new requirements.
		
		@param progress An object implementing the {@link ProgressReporter}
			interface
		@param rootLogin The root user login
		@param rootPassword The root user password
		@throws BaseException
	 */
	public static synchronized void updateDatabase(ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		if (progress != null) progress.display(0, "Updating database...");
		org.hibernate.Session session = null;
		try
		{
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			sc.login(rootLogin, rootPassword, null);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			int schemaVersion = getSchemaVersion(session);
			final int total_progress_steps = 1 + Install.NEW_SCHEMA_VERSION - schemaVersion;
			final float progress_step = 100F / total_progress_steps;
			float progress_current = progress_step;
			
			if (schemaVersion == 99)
			{
				// Final update to BASE 3
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 100...");
				schemaVersion = updateToSchemaVersion100(session);
				progress_current += progress_step;
				
				// Skipping directly to 102 since the updateToBase3() 
				// has already fixed those issues
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 102...");
				schemaVersion = setSchemaVersionInTransaction(session, 102);
				progress_current += 2 * progress_step;
			}

			// The 101 and 102 updates are only needed when updating from a BASE 3.0.0 installation
			if (schemaVersion < 101)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 101...");
				schemaVersion = updateToSchemaVersion101(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 102)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 102...");
				schemaVersion = updateToSchemaVersion102(session);
				progress_current += progress_step;
			}
			
			sc.logout();
			if (progress != null) progress.display(100, "Database updated successfully.");
		}
		catch (BaseException ex)
		{
			if (progress != null) progress.display(100, "Database update failed: " + ex.getMessage()+"\n");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	
	/**
		Verify the root user login and password. This method is called first 
		in the update procedure, which means that tables, etc. are usually
		still from the old version.
	*/
	static void verifyRootLoginAndPassword(String login, String password)
	{
	
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			int schemaVersion = 0;
			try
			{
				org.hibernate.Query query = HibernateUtil.createQuery(session, 
						"SELECT sv.schemaVersion FROM SchemaVersionData sv");
				schemaVersion = HibernateUtil.loadData(Integer.class, query);
			}
			catch (Throwable t)
			{}
			
			if (schemaVersion > 0)
			{
				org.hibernate.Query query = HibernateUtil.createQuery(session, 
					"SELECT usr.id, usr.login FROM UserData usr WHERE usr.systemId='" + User.ROOT + "'");
				Object[] user = (Object[])query.uniqueResult();
				if (user == null)
				{
					throw new LoginException("User ROOT not found");
				}
				if (!login.equals(user[1]))
				{
					throw new LoginException("User '" + login + "' is not the ROOT user");
				}
				
				int id = (Integer)user[0];
				
				if (schemaVersion > 90)
				{
					// BASE 3
					query = HibernateUtil.createQuery(session, 
						"SELECT cryptedPassword FROM PasswordData pwd WHERE pwd.id="+id);
					String cryptedPassword = HibernateUtil.loadData(String.class, query);
					if (!User.checkPassword(password, cryptedPassword))
					{
						throw new LoginException("Incorrect password for user '" + login + "'");
					}
				}
				else if (schemaVersion == 90)
				{
					// BASE 2.17
					query = HibernateUtil.createSqlQuery(session, 
						"SELECT [md5password] FROM [Passwords] WHERE [id]="+id);
					String md5Password = HibernateUtil.loadData(String.class, query);
					if (!MD5.getHashString(password).equals(md5Password))
					{
						throw new LoginException("Incorrect password for user '" + login + "'");
					}
				}
				// We let all other conditions pass, since they will 
				// usually generate an error later
			}
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Get the current schema version number as it is stored in the database.
		@return The schema version number, or 0 if no schem version is found
	*/
	private static int getSchemaVersion(org.hibernate.Session session)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		int schemaVersion = 0;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SCHEMA_VERSION");
			
			SchemaVersionData data = HibernateUtil.loadData(SchemaVersionData.class, query);
			schemaVersion = data == null ? 0 : data.getSchemaVersion();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("getSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
		Update the database with a new schema version number. This method should
		be called from an open transaction.
		
		@param schemaVersion The new schema version number
	 */
	private static void setSchemaVersion(org.hibernate.Session session, int schemaVersion)
		throws BaseException
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_SCHEMA_VERSION");
		query.setInteger("schemaVersion", schemaVersion);
		query.setInteger("build", Application.getBuild());
		HibernateUtil.executeUpdate(query);
	}
	
	/**
		Update the schema version version number using a new transaction.
		@param schemaVersion The new schema version number
	 */
	private static int setSchemaVersionInTransaction(org.hibernate.Session session, int schemaVersion)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			setSchemaVersion(session, schemaVersion);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("setSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Adjust the existing items in the database to be compatible with the latest mappings.
	 	No update to the schema version should be done here. The code must also  
	 	consider the case that the update failes at a later stage and that the same
	 	update is executed again.
	 	
	  	@param progress An object implementing the {@link ProgressReporter}
	 		interface
	  	@param rootLogin The root user login
	  	@param rootPassword The root user password
	  	@throws BaseException
	*/
	public static synchronized void adjustExistingItems(ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Session session = null;
		try
		{
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			sc.login(rootLogin, rootPassword, null);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			int schemaVersion = getSchemaVersion(session);
			tx = HibernateUtil.newTransaction(session);
			
			//	Commit the changes
			HibernateUtil.commit(tx);
			log.info("adjustExistingItems: OK");					
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (progress != null) progress.display(100, "The adjustment of the existing items failed: " + ex.getMessage()+"\n");
			log.info("adjustExistingItems: FAILED");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	
	/**
	 	Set subtypes for hybridizations, scans, 
	 	labeled extracts and labels moved from BASE 2.17
	 	
		@return The new schema version (=100)
	*/
	private static int updateToSchemaVersion100(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 100;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Labeled extract all have a non-null label_id column
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
				"SET_LABELEDEXTRACT_SUBTYPE");
			/*
				UPDATE BioMaterials SET subtype_id=? 
				WHERE label_id NOT IS NULL
			*/
			query.setInteger("subtype", SystemItems.getId(Extract.LABELED));
			HibernateUtil.executeUpdate(query);

			// All tags are labels
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"SET_SUBTYPE_ON_ALL", "Labels");
			// UPDATE Labels SET subtype_id=?
			query.setInteger("subtype", SystemItems.getId(Tag.LABEL));
			HibernateUtil.executeUpdate(query);
			
			// All physical bioassays are hybridizations
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"SET_SUBTYPE_ON_ALL", "PhysicalBioAssays");
			// UPDATE PhysicalBioassays SET subtype_id=?
			query.setInteger("subtype", SystemItems.getId(PhysicalBioAssay.HYBRIDIZATION));
			HibernateUtil.executeUpdate(query);
			
			// All derived bioassays are scans
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"SET_SUBTYPE_ON_ALL", "DerivedBioAssays");
			// UPDATE DerivedBioAssays SET subtype_id=?
			query.setInteger("subtype", SystemItems.getId(DerivedBioAssay.SCAN));
			HibernateUtil.executeUpdate(query);
			
			// All file set members for derived bioassays are microarray images
			query = HibernateUtil.getPredefinedQuery(session, "GET_DATAFILETYPE_FOR_EXTERNAL_ID");
			query.setString("externalId", "tmp.image");
			DataFileTypeData tmpType = HibernateUtil.loadData(DataFileTypeData.class, query);
			if (tmpType != null)
			{
				query.setString("externalId", DataFileType.MICROARRAY_IMAGE);
				DataFileTypeData imgType = HibernateUtil.loadData(DataFileTypeData.class, query);
				query = HibernateUtil.getPredefinedQuery(session, "UPDATE_FILESETMEMBER_FILETYPE");
				// UPDATE FileSetMemberData mbr SET mbr.dataFileType = :newType WHERE mbr.dataFileType = :oldType
				query.setInteger("newType", imgType.getId());
				query.setInteger("oldType", tmpType.getId());
				query.executeUpdate();

				// Remove the temporary data file type
				session.delete(tmpType);
			}
						
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion100: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion100: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Remove duplicate file set members.
	 	
		@return The new schema version (=101)
	*/
	private static int updateToSchemaVersion101(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 101;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT fsm FROM FileSetMemberData fsm");
			List<FileSetMemberData> members = HibernateUtil.loadList(FileSetMemberData.class, query, null);
			Set<String> allMembers = new HashSet<String>();
			for (FileSetMemberData member : members)
			{
				int fileSetId = member.getFileSet().getId();
				int fileTypeId = member.getDataFileType().getId();
				int fileId = member.getFile().getId();
				String unique = fileSetId + "." + fileTypeId + "." + fileId;
				if (!allMembers.add(unique))
				{
					// This combination has been seen before
					session.delete(member);
				}
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion101: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion101: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Add unique index on FileSetMembers(fileset_id,datafiletype_id,file_id)
	 	
		@return The new schema version (=102)
	*/
	private static int updateToSchemaVersion102(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 102;
		try
		{
	
			// Try to drop just to make sure that we don't create duplicated indexes
			dropIndex(session, true, "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			// Create the index
			createIndex(session, true, "FileSetMembers_uniquefile", "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);
	
			log.info("updateToSchemaVersion102: OK");
		}
		catch (SQLException ex)
		{
			log.error("updateToSchemaVersion102: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (BaseException ex)
		{
			log.error("updateToSchemaVersion102: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	// Item type codes for item types that was removed since BASE 2
	private static final int LABELEDEXTRACT = 204;
	private static final int HYBRIDIZATION = 261;
	private static final int SCAN = 262;
	private static final int IMAGE = 263;
	private static final int PROTOCOLTYPE = 102;
	private static final int HARDWARETYPE = 122;
	private static final int SOFTWARETYPE = 124;
	private static final int FILETYPE = 82;
	// Interval in milliseconds between adding a 'tick' to progress reporter
	private static final int TICK_INTERVAL = 5000;
	
	public static synchronized void updateToBase3(ProgressReporter progress, String rootLogin, String rootPassword)
		throws Exception
	{

		org.hibernate.Session session = null;
		org.hibernate.Query query = null;
		try
		{
			progress.display(0, "Updating to BASE 3.0...");
			//if (true) throw new RuntimeException("foo");
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			session = HibernateUtil.newSession();

			// #1641: Encrypt passwords and get rid of MD5 hashes
			progress.display(5, "--Encrypting existing passwords");
			encryptPasswords(session, progress);
			dropColumn(session, "Passwords", "md5password", progress);
			
			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			sc.login(rootLogin, rootPassword, "Upgrading to BASE 3");
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			// 1: copy data to new tables/columns
			// #1592: PluginDefinitions and PluginTypes jar_path should be copied to jar_file but only keep filename
			progress.display(10, "--Updating plug-in paths");
			copyJarPathToJarFile(session, "PluginDefinitions");
			copyJarPathToJarFile(session, "PluginTypes");
			cleanContextFromProperty(session, Item.PLUGINDEFINITION, "jarPath", "jarFile", false);
			cleanContextFromProperty(session, Item.PLUGINTYPE, "jarPath", "jarFile", false);
	
			// #1598: Four properties removed from DataFileType
			cleanContextFromProperty(session, Item.DATAFILETYPE, "validatorClass", null, true);
			cleanContextFromProperty(session, Item.DATAFILETYPE, "validatorJarPath", null, true);
			cleanContextFromProperty(session, Item.DATAFILETYPE, "metadataReaderClass", null, true);
			cleanContextFromProperty(session, Item.DATAFILETYPE, "metadataReaderJarPath", null, true);

			// #1604
			progress.display(15, "--Removing unique constraint on FileSetMembers");
			dropIndex(session, true, "FileSetMembers", "fileset_id", "datafiletype_id");
			createIndex(session, true, "FileSetMembers_uniquefile", "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			fixPlatformFileTypes(session);
			
			// #1153
			progress.display(20, "--Re-linking parent biomaterials");
			fixBioMaterialParents(session, progress);

			progress.display(25, "--Converting labeled extracts to extracts");
			createIndex(session, false, "ItemValues_item_id", "ItemValues", "data_class_id");
			createIndex(session, false, "AnyToAny_tmp", "AnyToAny", "from_id", "from_type");
			fixLabeledExtracts(session, progress);
			removeContext(session, LABELEDEXTRACT);
			
			progress.display(30, "--Converting hybridizations to physical bioassays");
			Map<Integer, Integer> hybMap = copyHybridizations(session, progress);
			cleanContextFromProperty(session, Item.ARRAYSLIDE, "hybridization", "physicalBioAssay", false);			
			removeContext(session, HYBRIDIZATION);
			
			progress.display(35, "--Converting scans to derived bioassays");
			Map<Integer, Integer> scanMap = copyScans(session, hybMap, progress);
			removeContext(session, SCAN);
			removeContext(session, IMAGE);
			
			progress.display(40, "--Re-linking rawbioassays to parents");
			fixRawBioAssays(session, scanMap, progress);
			cleanContextFromProperty(session, Item.RAWBIOASSAY, "scan", "parentBioAssay", false);
			
			// #1597
			progress.display(45, "--Converting protocol types to item subtypes");
			copyProtocolTypes(session, progress);
			cleanContextFromProperty(session, Item.PROTOCOL, "protocolType", "itemSubtype", true);
			cleanContextFromProperty(session, Item.PLATEEVENTTYPE, "protocolType", "protocolType", true);
			removeContext(session, PROTOCOLTYPE);

			progress.display(50, "--Converting software types to item subtypes");
			copySoftwareTypes(session, progress);
			cleanContextFromProperty(session, Item.SOFTWARE, "softwareType", "itemSubtype", true);
			removeContext(session, SOFTWARETYPE);

			progress.display(55, "--Converting hardware types to item subtypes");
			copyHardwareTypes(session, progress);
			cleanContextFromProperty(session, Item.HARDWARE, "hardwareType", "itemSubtype", true);
			removeContext(session, HARDWARETYPE);

			progress.display(60, "--Converting file types to item subtypes");
			copyFileTypes(session, progress);
			cleanContextFromProperty(session, Item.FILE, "fileType", "itemSubtype", true);
			cleanContextFromProperty(session, Item.MIMETYPE, "fileType", "fileType", true);
			cleanContextFromProperty(session, Item.DATAFILETYPE, "genericType", "genericType", true);
			removeContext(session, FILETYPE);
		
			progress.display(65, "--Re-linking project default items");
			fixProjectDefaultItems(session, progress);

			progress.display(70, "--Disabling all plug-ins");
			disablePlugins(session);
			
			// 2: remove unused columns/tables
			progress.display(80, "--Dropping old database objects");
			dropIndex(session, false, "ItemValues", "data_class_id");
			dropIndex(session, false, "AnyToAny", "from_id", "from_type");
			// #1592
			dropColumn(session, "PluginDefinitions", "jar_path", progress);
			dropColumn(session, "PluginTypes", "jar_path", progress);
			dropColumn(session, "JobAgentSettings", "jar_path", progress);
			// #1589
			dropColumn(session, "ArrayDesigns", "affy_chip", progress);
			// #1598
			dropColumn(session, "DataFileTypes", "validator_class", progress);
			dropColumn(session, "DataFileTypes", "validator_jarpath", progress);
			dropColumn(session, "DataFileTypes", "extractor_class", progress);
			dropColumn(session, "DataFileTypes", "extractor_jarpath", progress);
			
			// #1597
			dropColumn(session, "Files", "filetype_id", progress);
			dropColumn(session, "MimeTypes", "filetype_id", progress);
			dropColumn(session, "DataFileTypes", "filetype_id", progress);
			dropColumn(session, "Protocols", "protocoltype_id", progress);
			dropColumn(session, "PlateEventTypes", "protocoltype_id", progress);
			dropColumn(session, "Hardware", "hardwaretype_id", progress);
			dropColumn(session, "Software", "softwaretype_id", progress);
			dropTable(session, "FileTypes", progress);
			dropTable(session, "ProtocolTypes", progress);
			dropTable(session, "HardwareTypes", progress);
			dropTable(session, "SoftwareTypes", progress);
			
			// #1153
			dropColumn(session, "RawBioAssays", "scan_id", progress);
			dropColumn(session, "RawBioAssays", "array_num", progress);
			dropColumn(session, "BioMaterials", "pooled", progress);
			dropColumn(session, "BioMaterialEvents", "hybridization_id", progress);
			dropTable(session, "Images", progress);
			dropTable(session, "Scans", progress);
			dropTable(session, "Hybridizations", progress);
			dropTable(session, "BioMaterialEventSources", progress);
			
			setSchemaVersionInTransaction(session, 99);
			sc.logout();

			log.info("updateToBase3: OK");
		}
		catch (Exception ex)
		{
			progress.display(100, "Update to BASE 3 failed: " + ex.getMessage()+"\n");
			log.info("updateToBase3: FAILED");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}

	@SuppressWarnings("unchecked")
	private static void encryptPasswords(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;

			query = HibernateUtil.getPredefinedSQLQuery(session, "GET_MD5_PASSWORDS");
			/*
				SELECT id, md5password FROM Passwords
			*/
			List<Object[]> result = query.list();
			query = HibernateUtil.getPredefinedSQLQuery(session, "SET_ENCRYPTED_PASSWORD");
			/*
				UPDATE Passwords SET crypted_password = :cryptedPassword WHERE id = :id
			*/
			for (Object[] o : result)
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				int id = (Integer)o[0];
				String md5Password = Values.getStringOrNull((String)o[1]);
				if (md5Password != null)
				{
					String cryptedPassword = BCrypt.hashpw(md5Password, BCrypt.gensalt());
					log.debug("Updating password: id=" + id + "; md5Password=" + md5Password + "; crypted=" + cryptedPassword);
					query.setInteger("id", id);
					query.setString("cryptedPassword", cryptedPassword);
					query.executeUpdate();
				}
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}
	
	
	@SuppressWarnings("unchecked")
	private static void copyJarPathToJarFile(org.hibernate.Session session, String tableName)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			log.debug("Copy JAR path in table: " + tableName);

			query = HibernateUtil.getPredefinedSQLQuery(session, "GET_JAR_PATH_FROM_TABLE", tableName);
			List<Object[]> result = query.list();
			query = HibernateUtil.getPredefinedSQLQuery(session, "SET_JAR_FILE_IN_TABLE", tableName);
			for (Object[] o : result)
			{
				int id = (Integer)o[0];
				String jarPath = (String)o[1];
				int lastSlash = jarPath.lastIndexOf('/');
				String jarFile = jarPath.substring(lastSlash+1);
				log.debug("Setting JAR file: " + jarFile + "; id=" + id);
				query.setInteger("id", id);
				query.setString("file", jarFile);
				query.executeUpdate();
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}

	private static TableInfo getTableInfo(org.hibernate.Session session, final String tableName)
		throws SQLException
	{
		TableInfo info = HibernateUtil.doJdbcWork(session, 
		
				new JdbcWork<TableInfo>()
				{

					private TableInfo info;
					@Override
					public void execute(Connection c) 
						throws SQLException
					{
						DatabaseMetaData metaData = c.getMetaData();
						Table t = new Table(tableName);
						info = new TableInfo(t, metaData);
						
					}

					@Override
					public TableInfo getResult()
					{
						return info;
					}}
		
		);
		return info;
	}
	
	private static void dropColumn(org.hibernate.Session session, String tableName, String columnName, ProgressReporter progress)
		throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			DbEngine engine = HibernateUtil.getDbEngine();
			
			log.debug("Dropping column: " + tableName + "." + columnName);
			tx = HibernateUtil.newTransaction(session);
			
			// Drop foreign keys where this column is used
			TableInfo info = getTableInfo(session, tableName);
			for (ForeignKeyInfo fk : info.getForeignKeys())
			{
				if (fk.getFkColumns().contains(columnName))
				{
					log.debug("Dropping foreign key: " + tableName + "." + fk.getName() + "=" + fk.getFkColumns());
					String sql = engine.getDropForeignKeySql(null, null, tableName, fk.getName());
					query = HibernateUtil.createSqlQuery(session, sql);
					query.executeUpdate();
				}
			}
			
			progress.display(90, "  --alter table " + tableName + " drop column " + columnName);
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_COLUMN", tableName, columnName);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void dropIndex(org.hibernate.Session session, boolean unique, String tableName, String... columnNames)
			throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Dropping index: " + tableName + ".[" + Arrays.toString(columnNames) + "]");
			tx = HibernateUtil.newTransaction(session);
			
			TableInfo info = getTableInfo(session, tableName);
			String indexName = info.findIndexName(null, new HashSet<String>(Arrays.asList(columnNames)));
			if (indexName != null)
			{
				DbEngine engine = HibernateUtil.getDbEngine();
				String sql = engine.getDropIndexSql(null, null, tableName, indexName, unique);
				query = HibernateUtil.createSqlQuery(session, sql);
				query.executeUpdate();
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void createIndex(org.hibernate.Session session, boolean unique, String indexName, String tableName, String... columnNames)
			throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Creating index: " + tableName + ".[" + Arrays.toString(columnNames) + "]");
			tx = HibernateUtil.newTransaction(session);
			DbEngine engine = HibernateUtil.getDbEngine();
			String sql = engine.getCreateIndexSql(null, null, tableName, indexName, new LinkedHashSet<String>(Arrays.asList(columnNames)), unique);
			query = HibernateUtil.createSqlQuery(session, sql);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	
	private static void dropTable(org.hibernate.Session session, String tableName, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Dropping table: " + tableName);
			progress.display(90, "  --drop table " + tableName);
			tx = HibernateUtil.newTransaction(session);
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_TABLE", tableName);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}
	
	private static void disablePlugins(org.hibernate.Session session)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Disabling all plug-ins");
			tx = HibernateUtil.newTransaction(session);
			
			// Disable all plug-ins
			query = HibernateUtil.getPredefinedQuery(session, "DISABLE_ALL_PLUGINS");
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}

	}
	
	/**
		Remove context-related information for the given item type.
	*/
	private static void removeContext(org.hibernate.Session session, int itemType)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Removing context for item: " + itemType);
			tx = HibernateUtil.newTransaction(session);
			
			query = HibernateUtil.createQuery(session, "SELECT ctx FROM ContextData ctx WHERE ctx.itemType = :itemType");
			query.setInteger("itemType", itemType);
			for (ContextData ctx : HibernateUtil.loadList(ContextData.class, query, null))
			{
				session.delete(ctx);
			}

			// DELETE FROM PluginDefinitionGuiContexts WHERE item_type=?
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PLUGINGUICONTEXTS_FOR_ITEMTYPE");
			query.setInteger("itemType", itemType);
			query.executeUpdate();
			
			log.debug("Removing role key for item: " + itemType);
			// DELETE FROM PluginKeys WHERE item_type=?
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PLUGINKEYS_FOR_ITEMTYPE");
			query.setInteger("itemType", itemType);
			query.executeUpdate();

			query = HibernateUtil.createQuery(session, "SELECT key FROM RoleKeyData key WHERE key.itemType = :itemType");
			query.setInteger("itemType", itemType);
			for (RoleKeyData key : HibernateUtil.loadList(RoleKeyData.class, query, null))
			{
				session.delete(key);
			}
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}
	
	private static void cleanContextFromProperty(org.hibernate.Session session, Item itemType, 
		String propertyName, String replacementPropertyName, boolean forceRemoveFilter)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			log.debug("Cleaning context for item: " + itemType + "; property=" + propertyName + "; replacement=" + replacementPropertyName);
			tx = HibernateUtil.newTransaction(session);
			query = HibernateUtil.createQuery(session, "select ctx from ContextData ctx where ctx.itemType = " + itemType.getValue());
			List<ContextData> contexts = HibernateUtil.loadList(ContextData.class, query, null);
			
			for (ContextData ctx : contexts)
			{
				// Sort property
				if (ctx.getSortProperty() != null)
				{
					ctx.setSortProperty(replaceInCommaSeparatedList(ctx.getSortProperty(), propertyName, replacementPropertyName));
				}
				
				// "columns" settings
				Map<String, String> settings = ctx.getSettings();
				String columns = settings.get("columns");
				if (columns != null)
				{
					columns = replaceInCommaSeparatedList(columns, propertyName, replacementPropertyName);
					if (columns == null)
					{
						settings.remove("columns");
					}
					else
					{
						settings.put("columns", columns);
					}
				}
					
				// Filter
				Map<String, PropertyFilterData> filters = ctx.getPropertyFilters();
				PropertyFilterData propertyFilter = filters.remove(propertyName);
				if (propertyFilter != null && replacementPropertyName != null && !forceRemoveFilter)
				{
					// Replace with new property
					filters.put(replacementPropertyName, propertyFilter);
				}
				
			}
			
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}
	
	
	private static String replaceInCommaSeparatedList(String list, String find, String replacement)
	{
		List<String> elements = new ArrayList<String>(Arrays.asList(list.split(",")));
		String result = list;
		
		Iterator<String> it = elements.iterator();
		boolean hasMatched = false;
		int index = 0;
		while (it.hasNext())
		{
			String value = it.next();
			if (value != null && value.startsWith(find))
			{
				// A match, replace or remove
				hasMatched = true;
				if (replacement != null)
				{
					elements.set(index, replacement + value.substring(find.length()));
				}
				else
				{
					it.remove();
				}
			}
		}
		if (hasMatched)
		{
			if (elements.size() == 0)
			{
				result = null;
			}
			else
			{
				result = Values.getString(elements, ",", true);
			}
		}
		return result;
	}
	
	@SuppressWarnings("unchecked")
	private static <T> T load(org.hibernate.Session session, Class<T> clazz, Object id)
	{
		if (id == null) return null;
		return (T)session.load(clazz, (Integer)id);
	}
	
	private static void fixLabeledExtracts(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Query query = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			query = HibernateUtil.getPredefinedSQLQuery(session, "CONVERT_LABELEDEXTRACTS_TO_EXTRACTS");
			query.executeUpdate();
			
			fixSoftLinks(session, LABELEDEXTRACT, Item.EXTRACT.getValue(),  
				"net.sf.basedb.core.data.LabeledExtractData",
				"net.sf.basedb.core.data.ExtractData",
				null, null, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}
	
	/**
		Copy from 'Hybridizations' table to 'PhysicalBioAssays' table. 
		Fix all "soft links" related to "HYBRIDIZATION", so that they
		now use "PHYSICALBIOASSAY" instead.
		@return A Map with oldId -> newId for hybridizations
	*/
	private static Map<Integer, Integer> copyHybridizations(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		Map<Integer, Integer> annotationSetMap = new HashMap<Integer, Integer>();
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query hybQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_HYBRIDIZATIONS");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = hybQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				Integer hybId = (Integer)row[0];
				PhysicalBioAssayData bioAssay = new PhysicalBioAssayData();
				bioAssay.setArraySlide(load(session, ArraySlideData.class, row[2]));
				bioAssay.setSize((Integer)row[3]);
				AnnotationSetData annotationSet = load(session, AnnotationSetData.class, row[4]);
				bioAssay.setAnnotationSet(annotationSet);
				bioAssay.setName((String)row[5]);
				bioAssay.setDescription((String)row[6]);
				bioAssay.setRemoved((Boolean)row[7]);
				bioAssay.setItemKey(load(session, ItemKeyData.class, row[8]));
				bioAssay.setProjectKey(load(session, ProjectKeyData.class, row[9]));
				bioAssay.setOwner(load(session, UserData.class, row[10]));
				
				session.save(bioAssay);
				
				idMap.put(hybId, bioAssay.getId());
				if (annotationSet != null)
				{
					annotationSetMap.put(bioAssay.getId(), annotationSet.getId());
				}
			}
			// Ensure all new physical bioassays are in the database
			session.flush();
			session.clear();
			
			org.hibernate.Query pbaQuery =
					HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_BIOASSAYID_FROM_HYBID");
				// UPDATE BioMaterialEvents 
				// SET physicalbioassay_id = ? WHERE hybridization_id=?
			for (Map.Entry<Integer, Integer> entry : idMap.entrySet())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				pbaQuery.setInteger("hybId", entry.getKey());
				pbaQuery.setInteger("bioAssayId", entry.getValue());
				pbaQuery.executeUpdate();
			}
			
			fixSoftLinks(session, HYBRIDIZATION, Item.PHYSICALBIOASSAY.getValue(),  
				"net.sf.basedb.core.data.HybridizationData",
				"net.sf.basedb.core.data.PhysicalBioAssayData",
				idMap, annotationSetMap, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		return idMap;
	}
	
	/**
		Copy from 'Scans' table to 'DerivedBioAssays' table. 
		Fix all "soft links" related to "SCAN", so that they
		now use "DERIVEDBIOASSAY" instead. Copy from 'Images'
		to `FileSetMembers`. Use a temporary file type that
		is later (in updateToSchemaVersion100) changed to the
		"Microarray image" type.
		
		@param hybMap A Map with oldId -> newId for hybridizations
		@return A Map with oldId -> newId for scans
	*/
	private static Map<Integer, Integer> copyScans(org.hibernate.Session session, Map<Integer, Integer> hybMap, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		Map<Integer, Integer> annotationSetMap = new HashMap<Integer, Integer>();
		try
		{
			tx = HibernateUtil.newTransaction(session);
			DataFileTypeData fileType = null;
			org.hibernate.Query scanQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_SCANS");
			org.hibernate.Query imageQuery =
				HibernateUtil.getPredefinedSQLQuery(session, "GET_IMAGES");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = scanQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				Integer scanId = (Integer)row[0];
				DerivedBioAssayData bioAssay = new DerivedBioAssayData();
				bioAssay.setEntryDate((Date)row[2]);
				bioAssay.setPhysicalBioAssay(load(session, PhysicalBioAssayData.class, hybMap.get(row[3])));
				bioAssay.setHardware(load(session, HardwareData.class, row[4]));
				bioAssay.setProtocol(load(session, ProtocolData.class, row[5]));
				AnnotationSetData annotationSet = load(session, AnnotationSetData.class, row[6]);
				bioAssay.setAnnotationSet(annotationSet);
				bioAssay.setName((String)row[7]);
				bioAssay.setDescription((String)row[8]);
				bioAssay.setRemoved((Boolean)row[9]);
				bioAssay.setItemKey(load(session, ItemKeyData.class, row[10]));
				bioAssay.setProjectKey(load(session, ProjectKeyData.class, row[11]));
				bioAssay.setOwner(load(session, UserData.class, row[12]));
				
				session.save(bioAssay);
				idMap.put(scanId, bioAssay.getId());
				if (annotationSet != null)
				{
					annotationSetMap.put(bioAssay.getId(), annotationSet.getId());
				}
				
				// Images
				imageQuery.setInteger("scanId", scanId);
				Iterator it2 = imageQuery.list().iterator();
				FileSetData fileSet = null;
				
				// Keep track of duplicate files (since more than one image may reference the same fileId)
				Set<Integer> files = new HashSet<Integer>();
				
				while (it2.hasNext())
				{
					Object[] imgRow = (Object[])it2.next();
					Integer fileId = (Integer)imgRow[2];
					if (fileId != null && files.add(fileId))
					{
						if (fileSet == null)
						{
							fileSet = new FileSetData();
							fileSet.setItemType(Item.DERIVEDBIOASSAY.getValue());
							session.save(fileSet);
							bioAssay.setFileSet(fileSet);
						}
						if (fileType == null)
						{
							// Create a temporary file type, which will be removed
							// once the install has created the correct one
							fileType = new DataFileTypeData();
							fileType.setName("Temporary file type for scanned images");
							fileType.setExternalId("tmp.image");
							fileType.setItemType(Item.DERIVEDBIOASSAY.getValue());
							session.save(fileType);
						}
						
						FileSetMemberData member = new FileSetMemberData();
						member.setFile(load(session, FileData.class, fileId));
						member.setFileSet(fileSet);
						member.setDataFileType(fileType);
						session.save(member);
					}
				}
			}
			// Ensure all new derived bioassays, etc. are in the database
			session.flush();
			session.clear();
			
			fixSoftLinks(session, SCAN, Item.DERIVEDBIOASSAY.getValue(),  
				"net.sf.basedb.core.data.ScanData",
				"net.sf.basedb.core.data.DerivedBioAssayData",
				idMap, annotationSetMap, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		return idMap;
	}

	/**
		Find the correct parent derived bioassay and extract for 
		each raw bioassay using the old scan_id column. 
		
		@param scanMap A Map with oldId -> newId for scans
	*/
	private static void fixRawBioAssays(org.hibernate.Session session, Map<Integer, Integer> scanMap, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		try
		{
			tx = HibernateUtil.newTransaction(session);
			DataFileTypeData fileType = null;
			org.hibernate.Query rawQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_RAWBIOASSAYS");
			org.hibernate.Query extractQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "FIND_EXTRACT_FOR_RAWBIOASSAY_POSITION");
			org.hibernate.Query setQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_RAWBIOASSAY");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = rawQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				Integer rawId = (Integer)row[0];
				Integer scanId = (Integer)row[1];
				Integer arrayNum = (Integer)row[2];

				Integer parentId = scanMap.get(scanId);
				Integer extractId = null;
				
				extractQuery.setInteger("derivedBioAssayId", parentId);
				extractQuery.setInteger("position", arrayNum);
				List<Integer> extracts = HibernateUtil.loadList(Integer.class, extractQuery, null);
				if (extracts.size() == 1) extractId = extracts.get(0);
				
				setQuery.setInteger("rawBioAssayId", rawId);
				setQuery.setInteger("parentBioAssayId", parentId);
				setQuery.setParameter("extractId", extractId, Type.INT.getTypeWrapper().getHibernateType());
				setQuery.executeUpdate();
			}
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	
	/**
		Copy from 'ProtocolTypes' table to 'ItemSubtypes' table. Fix
		"hard links" in 'Protocols' and 'PlateEventTypes' tables.
		Fix all "soft links" related to "PROTOCOLTYPE", so that they
		now use "ITEMSUBTYPE" instead.
		@return A Map with oldId -> newId for protocoltypes
	*/
	private static Map<Integer, Integer> copyProtocolTypes(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		Set<String> names = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query typeQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_PROTOCOLTYPES");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = typeQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				Integer typeId = (Integer)row[0];
				ItemSubtypeData subtype = new ItemSubtypeData();
				subtype.setItemType(Item.PROTOCOL.getValue());
				subtype.setEntryDate((Date)row[2]);
				subtype.setName(ensureUnique((String)row[3], names));
				subtype.setDescription((String)row[4]);
				subtype.setRemoved((Boolean)row[5]);
				subtype.setSystemId((String)row[6]);
				session.save(subtype);
				
				idMap.put(typeId, subtype.getId());
			}
			
			// Ensure all new subtypes are in the database
			session.flush();
			session.clear();
			
			org.hibernate.Query protocolQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "Protocols", "subtype_id", "protocoltype_id");
			// UPDATE Protocols SET subtype_id = ? WHERE protocoltype_id=?
			org.hibernate.Query plateEventTypesQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "PlateEventTypes", "protocolsubtype_id", "protocoltype_id");
			// UPDATE PlateEventTypes SET protocolsubtype_id = ? WHERE protocoltype_id=?
			for (Map.Entry<Integer, Integer> entry : idMap.entrySet())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				int oldId = entry.getKey();
				int newId = entry.getValue();
				
				// Protocols.subtype_id
				protocolQuery.setInteger("oldTypeId", oldId);
				protocolQuery.setInteger("subtypeId", newId);
				protocolQuery.executeUpdate();
				
				// PlateEventTypes.subtype_id
				plateEventTypesQuery.setInteger("oldTypeId", oldId);
				plateEventTypesQuery.setInteger("subtypeId", newId);
				plateEventTypesQuery.executeUpdate();
			}

			
			fixSoftLinks(session, PROTOCOLTYPE, Item.ITEMSUBTYPE.getValue(),  
				"net.sf.basedb.core.data.ProtocolTypeData",
				"net.sf.basedb.core.data.ItemSubtypeData",
				idMap, null, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		return idMap;
	}

	/**
		Copy from 'HardwareTypes' table to 'ItemSubtypes' table. Fix
		"hard links" in 'Hardware' table.
		Fix all "soft links" related to "HARDWARETYPE", so that they
		now use "ITEMSUBTYPE" instead.
		@return A Map with oldId -> newId for hardwaretypes
	*/
	private static Map<Integer, Integer> copyHardwareTypes(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		Set<String> names = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query typeQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_HARDWARETYPES");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = typeQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				Integer typeId = (Integer)row[0];
				ItemSubtypeData subtype = new ItemSubtypeData();
				subtype.setItemType(Item.HARDWARE.getValue());
				subtype.setEntryDate((Date)row[2]);
				subtype.setName(ensureUnique((String)row[3], names));
				subtype.setDescription((String)row[4]);
				subtype.setRemoved((Boolean)row[5]);
				subtype.setSystemId((String)row[6]);
				session.save(subtype);
				
				idMap.put(typeId, subtype.getId());
			}
			
			// Ensure all new subtypes are in the database
			session.flush();
			session.clear();
			
			org.hibernate.Query hardwareQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "Hardware", "subtype_id", "hardwaretype_id");
			// UPDATE Hardware SET subtype_id = ? WHERE hardwaretype_id=?
			for (Map.Entry<Integer, Integer> entry : idMap.entrySet())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				int oldId = entry.getKey();
				int newId = entry.getValue();
				
				// Hardware.subtype_id
				hardwareQuery.setInteger("oldTypeId", oldId);
				hardwareQuery.setInteger("subtypeId", newId);
				hardwareQuery.executeUpdate();
			}
			
			fixSoftLinks(session, HARDWARETYPE, Item.ITEMSUBTYPE.getValue(),  
				"net.sf.basedb.core.data.HardwareTypeData",
				"net.sf.basedb.core.data.ItemSubtypeData",
				idMap, null, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		return idMap;
	}

	/**
		Copy from 'FileTypes' table to 'ItemSubtypes' table. Fix
		"hard links" in 'Files', 'MimeTypes' and 'DataFileTypes' tables.
		Fix all "soft links" related to "FILETYPE", so that they
		now use "ITEMSUBTYPE" instead.
		@return A Map with oldId -> newId for file types
	*/
	private static Map<Integer, Integer> copyFileTypes(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		Set<String> names = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query typeQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_FILETYPES");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = typeQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				Integer typeId = (Integer)row[0];
				ItemSubtypeData subtype = new ItemSubtypeData();
				subtype.setItemType(Item.FILE.getValue());
				subtype.setEntryDate((Date)row[2]);
				subtype.setName(ensureUnique((String)row[3], names));
				subtype.setDescription((String)row[4]);
				subtype.setSystemId((String)row[5]);
				session.save(subtype);
				
				idMap.put(typeId, subtype.getId());
			}
			
			// Ensure all new subtypes are in the database
			session.flush();
			session.clear();

			org.hibernate.Query fileQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "Files", "subtype_id", "filetype_id");
			// UPDATE Files SET subtype_id = ? WHERE filetype_id=?
			org.hibernate.Query mimeTypeQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "MimeTypes", "filesubtype_id", "filetype_id");
			// UPDATE MimeTypes SET filesubtype_id = ? WHERE filetype_id=?
			org.hibernate.Query dataFileTypeQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "DataFileTypes", "filesubtype_id", "filetype_id");
			// UPDATE DataFileTypes SET filesubtype_id = ? WHERE filetype_id=?
			for (Map.Entry<Integer, Integer> entry : idMap.entrySet())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				int oldId = entry.getKey();
				int newId = entry.getValue();
				
				// Files.subtype_id
				fileQuery.setInteger("oldTypeId", oldId);
				fileQuery.setInteger("subtypeId", newId);
				fileQuery.executeUpdate();
				
				// MimeTypes.subtype_id
				mimeTypeQuery.setInteger("oldTypeId", oldId);
				mimeTypeQuery.setInteger("subtypeId", newId);
				mimeTypeQuery.executeUpdate();

				// DataFileTypes.subtype_id
				dataFileTypeQuery.setInteger("oldTypeId", oldId);
				dataFileTypeQuery.setInteger("subtypeId", newId);
				dataFileTypeQuery.executeUpdate();
			}

			
			fixSoftLinks(session, FILETYPE, Item.ITEMSUBTYPE.getValue(),  
				"net.sf.basedb.core.data.FileTypeData",
				"net.sf.basedb.core.data.ItemSubtypeData",
				idMap, null, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		return idMap;
	}

	
	/**
		Copy from 'SoftwareTypes' table to 'ItemSubtypes' table. Fix
		"hard links" in 'Software' table.
		Fix all "soft links" related to "SOFTWARETYPE", so that they
		now use "ITEMSUBTYPE" instead.
		@return A Map with oldId -> newId for softwaretypes
	*/
	private static Map<Integer, Integer> copySoftwareTypes(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
		Set<String> names = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query typeQuery = 
				HibernateUtil.getPredefinedSQLQuery(session, "GET_SOFTWARETYPES");
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			Iterator it = typeQuery.list().iterator();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				
				Object[] row = (Object[])it.next();
				Integer typeId = (Integer)row[0];
				ItemSubtypeData subtype = new ItemSubtypeData();
				subtype.setItemType(Item.SOFTWARE.getValue());
				subtype.setEntryDate((Date)row[2]);
				subtype.setName(ensureUnique((String)row[3], names));
				subtype.setDescription((String)row[4]);
				subtype.setSystemId((String)row[5]);
				session.save(subtype);
				
				idMap.put(typeId, subtype.getId());
			}
			
			// Ensure all new subtypes are in the database
			session.flush();
			session.clear();
			
			org.hibernate.Query softwareQuery = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_SUBTYPEID_FROM_OLDTYPEID", "Software", "subtype_id", "softwaretype_id");
			// UPDATE Software SET subtype_id = ? WHERE softwaretype_id=?
			for (Map.Entry<Integer, Integer> entry : idMap.entrySet())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				int oldId = entry.getKey();
				int newId = entry.getValue();
				
				// Software.subtype_id
				softwareQuery.setInteger("oldTypeId", oldId);
				softwareQuery.setInteger("subtypeId", newId);
				softwareQuery.executeUpdate();
			}

			fixSoftLinks(session, SOFTWARETYPE, Item.ITEMSUBTYPE.getValue(),  
				"net.sf.basedb.core.data.SoftwareTypeData",
				"net.sf.basedb.core.data.ItemSubtypeData",
				idMap, null, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		return idMap;
	}

	/**
		Set PlatformFileTypes.allow_multiple to false for all with a null value
	*/
	private static void fixPlatformFileTypes(org.hibernate.Session session)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			HibernateUtil.createQuery(session, "UPDATE PlatformFileTypeData SET allow_multiple=false").executeUpdate();
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
		
	}
	
	/**
		Copy from 'BioMaterialEventSources' to 'BioMaterialEventSources2'
		table.
		
		Set the parent type property on all biomaterials.
		
	*/
	private static void fixBioMaterialParents(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			// Find out how to generate ID:s in INSERT statements
			DbEngine engine = HibernateUtil.getDbEngine();
			PersistentClass pc = HibernateUtil.getClassMapping(BioMaterialEventSourceData.class.getName());
			String selectNextId = engine.getInsertAutoIncrementSql(pc);
			
			tx = HibernateUtil.newTransaction(session);

			/*
				The BioMaterialEventSources table contains parent
				links for all biomaterials except sample->biosource
				which is in the BioMaterials.parent_id column only.
				The update need to make sure that ALL parent
				links are in the BioMaterialEventSources2 table. Items
				with a single parent should have the BioMaterial.parent_id
				set and BioMaterial.parent_type must also be set if there is
				at least one parent and should be null if there are no parents.
			*/
			
			// Copy from BioMaterialEventSoruces to BioMaterialEventSources2
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
					"COPY_BIOMATERIALEVENTSOURCES", selectNextId);
			query.executeUpdate();
			
			// Copy from BioMaterial.parent_id to BioMaterialEventSources2 for Sample->BioSource
			query = HibernateUtil.getPredefinedSQLQuery(session, 
					"COPY_SAMPLE_BIOSOURCE_PARENT_LINK", selectNextId);
			query.executeUpdate();

			// Nullify parent_id and parent type for all biomaterials
			query = HibernateUtil.getPredefinedSQLQuery(session, "NULLIFY_ALL_BIOMATERIAL_PARENTS");
			query.executeUpdate();
			
			// Get all parent info from BioMaterialEventSources2
			query = HibernateUtil.getPredefinedSQLQuery(session, "GET_BIOMATERIAL_PARENT_INFO");
			org.hibernate.Query setQuery = HibernateUtil.getPredefinedSQLQuery(session,
					"SET_BIOMATERIAL_PARENT");
			Iterator it = query.list().iterator();
			Integer lastChildId = null;
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				Object[] row = (Object[])it.next();
				
				Integer parentId = (Integer)row[0];
				Integer discriminator = (Integer)row[1];
				Integer childId = (Integer)row[2];
				
				// The second time we see the same child, the parent_id column should be nullified
				if (childId.equals(lastChildId))
				{
					parentId = null;
				}
				lastChildId = childId;
				Item parentType = Item.EXTRACT;
				if (discriminator == 1)
				{
					parentType = Item.BIOSOURCE;
				}
				else if (discriminator == 2)
				{
					parentType = Item.SAMPLE;
				}
				
				setQuery.setParameter("parentId", parentId, Type.INT.getTypeWrapper().getHibernateType());
				setQuery.setInteger("parentType", parentType.getValue());
				setQuery.setInteger("childId", childId);
				setQuery.executeUpdate();
			}
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}
	
	/**
		Copy project default items from 'AnyToAny' to 'ItemParameterValues'.
	*/
	private static void fixProjectDefaultItems(org.hibernate.Session session, ProgressReporter progress)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			
			tx = HibernateUtil.newTransaction(session);
			
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"GET_PROJECT_DEFAULT_ANYTOANY");
			List<AnyToAnyData> oldDefaultItems = HibernateUtil.loadList(AnyToAnyData.class, query, null);
			
			long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
			for (AnyToAnyData link : oldDefaultItems)
			{
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				int projectId = link.getFromId();
				ProjectData project = load(session, ProjectData.class, projectId);
				ItemParameterValueData defaultItems = project.getDefaultItems();
				if (defaultItems == null) 
				{
					defaultItems = new ItemParameterValueData();
					project.setDefaultItems(defaultItems);
				}
				
				Item defaultType = Item.fromValue(link.getToType());
				defaultItems.getValues().add(load(session, defaultType.getDataClass(), link.getToId()));
				
				HibernateUtil.deleteData(session, link);
			}
						
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}

	
	/**
		Fix all "soft links" in the BASE database. Soft links are links that
		are an item_type + item_id pair (or a class name + id)
		@param oldType The old item type code
		@param newType The new item type code (must be different from the old or the update may
			be incorrect)
		@param idMap Map with old -> new id values (null or empty is accepted)
		@param annotationSetMap Map with new item id -> annotation set id (null or empty is accepted)
	*/
	private static void fixSoftLinks(org.hibernate.Session session, 
		int oldType, int newType, String oldDataClass, String newDataClass,
		Map<Integer, Integer> idMap, Map<Integer, Integer> annotationSetMap,
		ProgressReporter progress)
	{
		if (oldType == newType) 
		{
			throw new AssertionError("Can't fix links when oldType==newType: " + oldType);
		}
		
		String oldTypeS = Integer.toString(oldType);
		String newTypeS = Integer.toString(newType);

		long nextTick = System.currentTimeMillis()+TICK_INTERVAL; // Every ten seconds
		if (annotationSetMap != null && annotationSetMap.size() > 0)
		{
			// UPDATE AnnotationSets SET item_id=?, item_type={1} WHERE id=?
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_ANNOTATIONSET_ITEM", newTypeS);
			
			for (Map.Entry<Integer, Integer> entry : annotationSetMap.entrySet())
			{
				query.setInteger("id", entry.getValue());
				query.setInteger("newId", entry.getKey());
				query.executeUpdate();
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
			}
		}

		
		if (idMap != null && idMap.size() > 0)
		{
			// Queries that need to update ID of each individual item
			List<org.hibernate.Query> queries = new ArrayList<org.hibernate.Query>();
				
			// UPDATE AnyToAny SET from_id=?, from_type={2} WHERE from_id=? and from_type={1}
			queries.add(HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_ANYTOANY_FROMITEM", oldTypeS, newTypeS));
			
			// UPDATE AnyToAny SET to_id=?, to_type={2} WHERE to_id=? and to_type={1}
			queries.add(HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_ANYTOANY_TOITEM", oldTypeS, newTypeS));
	
			// UPDATE ChangeHistoryDetails SET item_id=?, item_type={2} WHERE item_id=? and item_type={1}
			queries.add(HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_CHANGEHISTORYDETAILS_ITEM", oldTypeS, newTypeS));
			
			// UPDATE ItemValues SET data_class_id=?, data_class=? WHERE data_class_id=? AND data_class=?
			org.hibernate.Query q = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_ITEMVALUES_ITEM");
			q.setString("oldClass", oldDataClass);
			q.setString("newClass", newDataClass);
			queries.add(q);
			
			for (Map.Entry<Integer, Integer> entry : idMap.entrySet())
			{
				int oldId = entry.getKey();
				int newId = entry.getValue();
				for (org.hibernate.Query query : queries)
				{
					query.setInteger("oldId", oldId);
					query.setInteger("newId", newId);
					query.executeUpdate();
				}
				
				if (System.currentTimeMillis() > nextTick)
				{
					progress.append(".");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
			}
		}
		else
		{
			// Assume 1-1 mapping on ID and modify the item type only
			
			// UPDATE AnyToAny SET from_type={2} WHERE from_type={1}
			HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
				"AnyToAny", "from_type", oldTypeS, newTypeS).executeUpdate();
			// UPDATE AnyToAny SET to_type={2} WHERE to_type={1}
			HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
				"AnyToAny", "to_type", oldTypeS, newTypeS).executeUpdate();
			// UPDATE ChangeHistoryDetails SET item_type={2} WHERE item_type={1}
			HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
				"ChangeHistoryDetails", "item_type", oldTypeS, newTypeS).executeUpdate();
			
			// UPDATE ItemValues SET data_class_id=?, data_class=? WHERE data_class_id=? AND data_class=?
			org.hibernate.Query q = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_ITEMVALUES_CLASS");
			q.setString("oldClass", oldDataClass);
			q.setString("newClass", newDataClass);
			q.executeUpdate();
		}

		// Queries that work on the item type only
		// UPDATE AnnotationSet SET item_type={2} WHERE item_type={1}
		HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
			"AnnotationSets", "item_type", oldTypeS, newTypeS).executeUpdate();
		// UPDATE AnnotationTypeItems SET item_type={2} WHERE item_type={1}
		HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
			"AnnotationTypeItems", "item_type", oldTypeS, newTypeS).executeUpdate();
		// UPDATE BioMaterialLists SET member_type={2} WHERE member_type={1}
		HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
			"BioMaterialLists", "member_type", oldTypeS, newTypeS).executeUpdate();
		// UPDATE BioPlateTypes SET biomaterial_type={2} WHERE biomaterial_type={1}
		HibernateUtil.getPredefinedSQLQuery(session, "UPDATE_ITEMTYPE", 
			"BioPlateTypes", "biomaterial_type", oldTypeS, newTypeS).executeUpdate();
		
	}
	
	/**
		Ensure that the given name is unique among the set of names.
		If the name doesn't exists in the set it is added to it and
		returned unmodified. If the name already exists we try
		to append (2), (3), ... and so on until an unused name is 
		found.
		@param name The original name
		@param names The set of already used names
		@return A unique name
	*/
	private static String ensureUnique(String name, Set<String> names)
	{
		String uniqueName = name;
		int i = 2;
		while (names.contains(uniqueName))
		{
			uniqueName = name + " (" + i + ")";
			i++;
		}
		names.add(uniqueName);
		return uniqueName;
	}
}
