/*
	$Id: Authenticator.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.authentication;

/**
	This interface defines the methods that needs to be implemented 
	by a class to be able to use it for external authentication against
	another database or directory service.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public interface Authenticator
{
	/**
		Initialises the object. A client application should call this
		method immediately after object creation with some initialisation
		parameters. The contents and syntax of the <code>settings</code>
		string should be defined by the implementing class. It may for
		example be the URL for a database connection,
		the path to a properties file or simply the host name of the
		directory service provider.

		@param settings A string containing initialisation parameters
		@throws AuthenticationException If the parameters are invalid
	*/
	public void init(String settings)
		throws AuthenticationException;

	/**
		Should return TRUE or FALSE depending on if the authentication
		server returns additional information about the user. 
	  	@return TRUE if additional information is returned, FALSE otherwise. 
		@see AuthenticationInformation
		@see #authenticate(String,String)
	*/
	public boolean supportsExtraInformation();
	
	/**
		Authenticates a user. If the authentication succeeds a
		{@link AuthenticationInformation} object
		should be returned. The only required information in this object is
		the {@link AuthenticationInformation#id} and
		{@link AuthenticationInformation#login} fields. The <code>id</code>
		may or may not be the same as the <code>login</code>, but must be a unique
		string identifying the user.
		<p>
		If the authentication fails an exception should be thrown. Note! An
		implementation may choose to throw a <code>AuthenticationException</code>
		for all errors, even if it is an unknown user or invalid password.

		@param login The login name for the user
		@param password The password for the user
		@throws UnknownLoginException This exception may be thrown if the
			server doesn't know about the user. It is also possible for an
			implmentation to throw an <code>AuthenticationException</code>
		@throws InvalidPasswordException This exception may be thrown if the
			password is invalid. It is also possible for an implementation to throw
			an <code>AuthenticationException</code>
		@throws AuthenticationException This exception is thrown if the
			authentication fails
		@return An <code>AuthenticationInformation</code> object
	*/
	public AuthenticationInformation authenticate(String login, String password)
		throws UnknownLoginException, InvalidPasswordException, AuthenticationException;

}
