/**
	$Id: BioPlateData.java 5818 2011-10-21 11:15:41Z nicklas $

	Copyright (C) 2008 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 	This class holds information about a bioplate
  	
  	@author martin
 	@version 2.10
 	@see net.sf.basedb.core.BioPlate
	@see <a href="../../../../../../html/developer/api/data_api.biomaterials.html">Developer documentation: Biomaterials LIMS</a>
	@base.modified $Date: 2008-11-27 16:33:55 +0100 (Thu, 27 Nov 2008) $
	@hibernate.class table="`BioPlates`" lazy="false"
 */
public class BioPlateData 
	extends AnnotatedData
{
	public BioPlateData()
	{}
	
	/**
		The maximum length of the external id that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = 255;
	private String externalId;
	/**
		Get the external id for the biomaterial
		@hibernate.property column="`external_id`" type="string" length="255" not-null="false"
	*/
	public String getExternalId()
	{
		return externalId;
	}
	public void setExternalId(String externalId)
	{
		this.externalId = externalId;
	}
	
	/**
		The maximum length of the barcode that can be stored in the database.
		@see #setBarcode(String)
	*/
	public static final int MAX_BARCODE_LENGTH = 255;
	private String barcode;
	/**
		Get the barcode for the bioplate. It is not required and doesn't have to be unique.
		@hibernate.property column="`barcode`" type="string" length="255" not-null="false" 
	*/
	public String getBarcode()
	{
		return barcode;
	}
	public void setBarcode(String barcode)
	{
		this.barcode = barcode;
	}
	
	private boolean destroyed;
	/**
		If the plate has been destroyed.
		@hibernate.property column="`destroyed`" type="boolean" not-null="true" 
		@since 2.16
	*/
	public boolean isDestroyed()
	{
		return destroyed;
	}
	public void setDestroyed(boolean destroyed)
	{
		this.destroyed = destroyed;
	}
	
	private PlateGeometryData plateGeometry;
	/**
		The geometry of this plate type.
		@hibernate.many-to-one column="`plategeometry_id`" not-null="true" outer-join="false" update="false"
	*/
	public PlateGeometryData getPlateGeometry()
	{
		return plateGeometry;
	}
	public void setPlateGeometry(PlateGeometryData plateGeometry)
	{
		this.plateGeometry = plateGeometry;
	}
	
	private HardwareData freezer;
	/**
		Get the hardware, ie. freezer where the plate is kept.
		@hibernate.many-to-one column="`hardware_id`" not-null="false" outer-join="false"
	*/
	public HardwareData getFreezer()
	{
		return freezer;
	}
	public void setFreezer(HardwareData freezer)
	{
		this.freezer = freezer;
	}
	
	private BioPlateTypeData bioPlateType;
	/**
		Get the type of the bioplate.
		@since 2.16
		@hibernate.many-to-one column="`bioplatetype_id`" not-null="true" 
			outer-join="false" update="false"
	*/
	public BioPlateTypeData getBioPlateType()
	{
		return bioPlateType;
	}
	/**
		@since 2.16
	*/
	public void setBioPlateType(BioPlateTypeData bioPlateType)
	{
		this.bioPlateType = bioPlateType;
	}
	
	private int freeWells;
	/**
		Get the number of wells that are free on the bioplate.
		@since 2.17
		@hibernate.property column="`free_wells`" type="int" not-null="true" 
	*/
	public int getFreeWells()
	{
		return freeWells;
	}
	public void setFreeWells(int freeWells)
	{
		this.freeWells = freeWells;
	}
	
	/**
		Update the number of free wells with the given difference
		@since 2.17
	*/
	public void updateFreeWells(int diff)
	{
		this.freeWells += diff;
	}
	
	
	private Map<PlateCoordinate, BioWellData> bioWells;
	/**
		The wells on this plate. This is the inverse end.
		@see BioWellData#getBioPlate()
		@hibernate.map lazy="true" inverse="true" cascade="all"
		@hibernate.collection-key column="`bioplate_id`"
		@hibernate.collection-composite-index class="net.sf.basedb.core.data.PlateCoordinate"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.BioWellData"
	*/
	public Map<PlateCoordinate, BioWellData> getBioWells()
	{
		if (bioWells == null)
		{
			PlateGeometryData pg = getPlateGeometry();
			if (pg != null)
			{
				int numWells = pg.getRows() * pg.getColumns();
				bioWells = new HashMap<PlateCoordinate, BioWellData>(numWells);
				for (int row = 0; row<pg.getRows(); row++)
				{
					for (int column=0; column<pg.getColumns(); column++)
					{
						BioWellData well = new BioWellData();
						well.setBioPlate(this);
						well.setRow(row);
						well.setColumn(column);
						bioWells.put(new PlateCoordinate(row, column), well);
					}
				}
			}
		}
		return bioWells;
	}
	void setBioWells(Map<PlateCoordinate, BioWellData> bioWells)
	{
		this.bioWells = bioWells;
	}
	
	private Set<BioPlateEventParticipantData> eventParticipants;
	/**
		This is the inverse end.
		@see BioPlateEventParticipantData#getBioPlate()
		@hibernate.set lazy="true" inverse="true" cascade="delete"
		@hibernate.collection-key column="`bioplate_id`"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.BioPlateEventParticipantData"
		@since 2.17
	*/
	Set<BioPlateEventParticipantData> getEventParticipants()
	{
		return eventParticipants;
	}
	void setEventParticipants(Set<BioPlateEventParticipantData> eventParticipants)
	{
		this.eventParticipants = eventParticipants;
	}

	
}