/**
	$Id: Signal.java 5399 2010-09-02 12:09:15Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
	Represents a signal that can be sent to a class implementing the
	{@link SignalTarget} interface. Signals must be registered before they
	can be used. Use {@link #getSignals()} to get a list of signals
	that have been registered. A signal can only be registered once.
	
	@author nicklas
	@version 2.6
	@base.modified $Date: 2010-09-02 14:09:15 +0200 (Thu, 02 Sep 2010) $
*/
public class Signal
{

	/**
		The ABORT signal, that requests that the receiver should
		clean up and die. This signal is typically sent to a plug-in
		when the user manually requests that a job is aborted. It is
		also sent when the system is shutting down to plug-ins that 
		doesn't support the {@link #SHUTDOWN} signal.
	*/
	public static final Signal ABORT = 
		registerSignal("ABORT", "Abort", 
			"Sending this signal will ask the receiver to abort " +
			"whatever it is doing, cleanup allocated resources and exit."
		);

	/**
		The SHUTDOWN signal, that requests that the receiver should
		clean up and die. When the system is up again the receiver
		may continue were it was stopped if supported. Otherwise, 
		this signal should be treated as {@link #ABORT}.
		@since 2.16
	*/
	public static final Signal SHUTDOWN = 
		registerSignal("SHUTDOWN", "Shutdown", 
			"Sending this signal indicates that the system is shutting " +
			"down. The receiver should stop whatever it is doing, " +
			"cleanup allocated resource and exit. When the system is back " +
			"up again, the reciever may continue working were it was stopped."
		);

	
	/**
		Holds all registered signals.
	*/
	private static Map<String, Signal> signals;
	
	/**
		Register a signal.
		
		@param id The ID of the signal. This must be unique. If another signal 
			with the same ID already exists, this method does nothing.
		@param name The short name of the signal
		@param description A description of the what the signal means and how 
			receivers should behave when the signal is sent to them
		@return The Signal object
	*/
	public static synchronized Signal registerSignal(String id, String name, String description)
	{
		if (signals == null) signals = new HashMap<String, Signal>();
		Signal s = signals.get(id);
		if (s == null)
		{
			s = new Signal(id, name, description);
			signals.put(id, s);
		}
		return s;
	}
	
	/**
		Get the signal with a given ID.
		@param id The signal ID.
		@return The Signal object, or null if no signal with the given ID
			has been registered
		@see #getSignals()
	*/
	public static Signal getSignal(String id)
	{
		return signals.get(id);
	}
	
	/**
		Get all registered signals. The returned collection can't be modified.
		@return A Collection with all registered signals
		@see #getSignal(String)
	*/
	public static Collection<Signal> getSignals()
	{
		return Collections.unmodifiableCollection(signals.values());
	}
	
	private final String id;
	private final String name;
	private final String description;
	
	private Signal(String id, String name, String description)
	{
		this.id = id;
		this.name = name;
		this.description = description;
	}
	
	/*
		From the Object class
		-------------------------------------------
	*/
	/**
		@return The name of the signal.
	*/
	@Override
	public String toString()
	{
		return name;
	}
	
	/**
		Two signals are equal if they have the same ID.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (!(other instanceof Signal)) return false;
		return this.id.equals(((Signal)other).id);
	}

	/**
		The hashcode for a signal is the same as the hashcode for
		the ID.
	*/
	@Override
	public int hashCode()
	{
		return id.hashCode();
	}
	// -------------------------------------------

	/**
		Get the signal ID. The ID must be unique among all
		registered signals.
	*/
	public String getId()
	{
		return id;
	}
	
	/**
		Get the short name of the signal.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Get the description of the what the signal means and how 
		receivers should behave when the signal is sent to them
	*/
	public String getDescription()
	{
		return description;
	}	
}
