/**
	$Id: AnnotationUtil.java 5749 2011-09-19 12:31:39Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.filter.Filter;

/**
	Utility class for working with annotations.

	@author Nicklas
	@version 2.9
	@base.modified $Date: 2011-09-19 14:31:39 +0200 (Mon, 19 Sep 2011) $
*/
public class AnnotationUtil
{
	
	/**
		Get all annotatable parent items for a given item. A {@link Filter}
		may be used to filter the returned collection. If a filter
		is specified only parent items for which the {@link Filter#evaluate(Object)}
		method returns true are included in the returned collection. Note
		that all parent items are loaded before the filter is evaluated
		and that the evaluation order is undefined.
		
		@param dc The DbControl to use for database access
		@param item The item to find the parents for
		@param filter An optional filter that can be used to filter 
			which parents to include
		@return A set containing the parent items
	*/
	public static Set<Annotatable> getAllAnnotatableParentItems(DbControl dc, 
		Annotatable item, Filter<Annotatable> filter)
	{
		Set<AnnotatableWrapper> processed = new HashSet<AnnotatableWrapper>();
		Queue<AnnotatableWrapper> unprocessed = new LinkedList<AnnotatableWrapper>();
		unprocessed.add(new AnnotatableWrapper(item, null));

		// As long as we have unprocessed entries...
		while (!unprocessed.isEmpty())
		{
			AnnotatableWrapper toProcess = unprocessed.remove();
			// ...mark it as processed...
			if (processed.add(toProcess))
			{
				// ...and queue the parents items for processing...
				unprocessed.addAll(toProcess.getAnnotatableParentWrappers());
			}
		}
		
		// All parents are now loaded...
		Set<Annotatable> allParents = new HashSet<Annotatable>(processed.size());
		for (AnnotatableWrapper wrapper : processed)
		{
			// ...copy all parents that passes the filter
			Annotatable parent = wrapper.getAnnotatable();
			if (filter == null || filter.evaluate(parent)) allParents.add(parent);
		}
		// Do not include the root item in the result
		allParents.remove(item);
		return allParents;
	}

	
	/**
		This wrapper class is needed because we need to keep
		track of the {@link RawBioAssay#getParentExtract()} property 
		until we have loaded the Extracts of a Physical Bioassay.
		<p>
		When we reach a {@link RawBioAssay} item, we remember the parent extract on
		that as we move up to DerivedBioAssaySet and PhysicalBioAssay.
		When the PhysicalBioAssay is reached we use the extract to call
		{@link PhysicalBioAssay#getAnnotatableParents(int, Extract)} so that we only
		get extracts that are linked with the correct raw bioassay.
	*/
	static class AnnotatableWrapper
	{
		/**
	 		Get the extract to use. This method returns:
	 		
	 		<ul>
	 		<li>If the item is a raw bioassay, {@link RawBioAssay#getParentExtract()}
	 		<li>If the item is a derived bioassay, extract if it is not null,
	 			otherwise {@link DerivedBioAssay#getExtract()}
	 		<li>If the item is a physical bioassay, the extract parameter
	 		<li>null in all other cases.
		*/
		private static Extract getExtract(Annotatable item, Extract extract)
		{
			Extract result = null;
			if (item instanceof RawBioAssay)
			{
				result = ((RawBioAssay)item).getParentExtract();
			}
			else if (item instanceof DerivedBioAssay)
			{
				if (extract != null)
				{
					result = extract;
				}
				else
				{
					result = ((DerivedBioAssay)item).getExtract();
				}
			}
			else if (item instanceof PhysicalBioAssay)
			{
				result = extract;
			}
			return result;
		}
		
		private final Annotatable item;
		private final Extract extract;

		AnnotatableWrapper(Annotatable item, Extract extract)
		{
			this.item = item;
			this.extract = getExtract(item, extract);
		}
		
		/**
			The annotatable that we are wrapping.
		*/
		Annotatable getAnnotatable()
		{
			return item;
		}
		
		/**
			Get all annotatable parents wrapped inside AnnotatableWrapper:s
		*/
		Set<AnnotatableWrapper> getAnnotatableParentWrappers() 
			throws BaseException
		{
			Set<Annotatable> parents = null;
			Set<AnnotatableWrapper> wrappedParents = Collections.emptySet();
			if (item instanceof PhysicalBioAssay)
			{
				parents = ((PhysicalBioAssay)item).getAnnotatableParents(0, extract);
			}
			else
			{
				parents = item.getAnnotatableParents();
			}
			if (parents != null)
			{
				wrappedParents = new HashSet<AnnotatableWrapper>(parents.size());
				for (Annotatable parent : parents)
				{
					wrappedParents.add(new AnnotatableWrapper(parent, extract));
				}
			}
			return wrappedParents;
		}
		
		/**
			A wrapper is equal to another if they references the same item
			and extract.
		*/
		@Override
		public boolean equals(Object o)
		{
			if (o == this) return true;
			if (o == null || o.getClass() != this.getClass()) return false;
			AnnotatableWrapper other = (AnnotatableWrapper)o;
			return this.item.equals(other.item) && EqualsHelper.equals(this.extract, other.extract);
		}

		@Override
		public int hashCode()
		{
			return item.hashCode() + EqualsHelper.hashCode(extract);
		}
	}
}
