/**
	$Id: ScanImporter.java 5734 2011-09-13 09:41:48Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

import java.util.Collections;
import java.util.List;
import java.util.Set;

/**
	Plug-in for importing scan items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2011-09-13 11:41:48 +0200 (Tue, 13 Sep 2011) $
	@deprecated In 3.0, use {@link DerivedBioAssayImporter} instead
*/
@Deprecated
public class ScanImporter
	extends AbstractItemImporter<DerivedBioAssay>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.DERIVEDBIOASSAY, GuiContext.Type.LIST));
	
	protected static final PluginParameter<String> hardwareColumnMapping = new PluginParameter<String>(
			"hardwareColumnMapping",
			"Scanner",
			"Mapping that picks the name or ID of the array slide scanner from the data columns. " +
			"The plug-in will first try to find a scanner with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Scanner\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> hybridizationColumnMapping = new PluginParameter<String>(
			"hybridizationColumnMapping",
			"Hybridization",
			"Mapping that picks the name or ID of the hybridization from the data columns. " +
			"The plug-in will first try to find a hybridization with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Hybridization\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> imageFileColumnMapping = new PluginParameter<String>(
			"imageFileColumnMapping",
			"Image file",
			"Mapping that picks the path of a image file from the columns. " +
			"The path can be an absolute path (starting with /) or relative " +
			"to the 'Data directory' parameter." +
			"The plug-in tries to find an existing image with the given file path and that belongs to current scan. "+
			"Example: \\Image file\\",
			optionalColumnMapping
			);
	
	public ScanImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper protocolMapper;
	private Mapper hardwareMapper;
	private Mapper hybridizationMapper;
	private Mapper imageFileMapper;
	private String imagesDirectory;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Adds column mappings for name, internalId, description,
		hybridization, protocol, hardware, image-file, image-name,
		image-format and image-description.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(hybridizationColumnMapping);
		parameters.add(protocolColumnMapping);
		parameters.add(hardwareColumnMapping);
		parameters.add(imageFileColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Add parameter for "Image directory" directory where image files may be located. 
	*/
	@Override
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(new PluginParameter<String>(
				"imagesDirectory",
				"Images directory",
				"A base directory were image files are located. Relative file paths are " +
				"resolved against this directory. If no directory is given, only absolute " +
				"paths (=paths starting with /) can be used.",
				new PathParameterType(Path.Type.DIRECTORY))
		);
	}

	
	/**
		Calls {@link DerivedBioAssay#getQuery()} and set a subtype filter for
		{@link DerivedBioAssay#SCAN}.
	*/
	@Override
	protected ItemQuery<DerivedBioAssay> createItemQuery()
	{
		ItemQuery<DerivedBioAssay> query = DerivedBioAssay.getQuery();
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property("itemSubtype"), 
					Expressions.integer(SystemItems.getId(DerivedBioAssay.SCAN))
				)
			);
		return query;
	}
		
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? DerivedBioAssay.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? DerivedBioAssay.MAX_DESCRIPTION_LENGTH : null, null);
		protocolMapper = getMapper(ffp, (String)job.getValue("protocolColumnMapping"), null, null);
		hardwareMapper = getMapper(ffp, (String)job.getValue("hardwareColumnMapping"), null, null);
		hybridizationMapper = getMapper(ffp, (String)job.getValue("hybridizationColumnMapping"), null, null);
		imageFileMapper = getMapper(ffp, (String)job.getValue("imageFileColumnMapping"), null, null);
		imagesDirectory = (String)job.getValue("imagesDirectory");
		if (imagesDirectory != null && !imagesDirectory.startsWith("/"))
		{
			imagesDirectory += "/";
		}
	}
	
	@Override
	protected DerivedBioAssay createItem(DbControl dc, FlatFileParser.Data data)
	{
		PhysicalBioAssay hyb = null;
		if (hybridizationMapper != null) 
		{
			hyb = findHybridization(dc, FallbackIdMethod.NAME_OR_ID, hybridizationMapper.getValue(data));
		}
		DerivedBioAssay scan = DerivedBioAssay.getNew(dc, hyb, job.getJob());
		scan.setItemSubtype(ItemSubtype.getById(dc, SystemItems.getId(DerivedBioAssay.SCAN)));
		updateItem(dc, scan, data);
		return scan;
	}
	
	@Override
	protected void updateItem(DbControl dc, DerivedBioAssay scan, FlatFileParser.Data data)
	{
		if (nameMapper != null) scan.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) scan.setDescription(descriptionMapper.getValue(data));
		if (protocolMapper != null) 
		{
			String nameOrId = protocolMapper.getValue(data);
			ItemSubtype type = ItemSubtype.getById(dc, SystemItems.getId(Protocol.SCANNING));
			Protocol protocol = findProtocol(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (nameOrId == null || protocol != null) scan.setProtocol(protocol);
		}
		if (hardwareMapper != null)
		{
			String nameOrId = hardwareMapper.getValue(data);
			ItemSubtype type = ItemSubtype.getById(dc, SystemItems.getId(Hardware.SCANNER));
			Hardware scanner = findHardware(dc, FallbackIdMethod.NAME_OR_ID, nameOrId, type);
			if (nameOrId == null || scanner != null) scan.setHardware(scanner);
		}
		updateMultiLineItem(dc, scan, data, 0);
	}
	
	@Override
	protected void updateMultiLineItem(DbControl dc, DerivedBioAssay scan, FlatFileParser.Data data, int multiLineNum)
	{

		// Get the imagefile
		if (imageFileMapper != null)
		{
			String filePath = imageFileMapper.getValue(data);
			if (!filePath.startsWith("/") && imagesDirectory != null)
			{
				filePath = imagesDirectory + filePath;
			}
			File imageFile = File.getByPath(dc, new Path(filePath, Path.Type.FILE), false);
			DataFileType imageType = DataFileType.getByExternalId(dc, DataFileType.MICROARRAY_IMAGE);
			setOrAddFile(scan, imageType, imageFile);
		}
	}
	// ------------------------------------
}
