/**
	$Id: AdminTest.java 5827 2011-10-26 10:51:52Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test.roles;

import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Group;
import net.sf.basedb.core.MultiPermissions;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.Quota;
import net.sf.basedb.core.Role;
import net.sf.basedb.core.SharedItem;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.User;
import net.sf.basedb.plugins.PlateFlatFileImporter;
import net.sf.basedb.plugins.PrintMapFlatFileImporter;
import net.sf.basedb.plugins.ReporterMapFlatFileImporter;
import net.sf.basedb.plugins.batchimport.ArrayBatchImporter;
import net.sf.basedb.plugins.batchimport.ArrayDesignImporter;
import net.sf.basedb.plugins.batchimport.ArraySlideImporter;
import net.sf.basedb.plugins.gtf.GtfReporterMapImporter;
import net.sf.basedb.test.AffymetrixData;
import net.sf.basedb.test.FileUtil;
import net.sf.basedb.test.MouseData;
import net.sf.basedb.test.PluginUtil;
import net.sf.basedb.test.SequenceData;
import net.sf.basedb.test.TestUtil;

/**
	This class contains the tests for the administrator created
	by the root user.
	<ol>
	<li>Create group: {@link #createGroup(DbControl)}
	<li>Create power user: {@link #createPowerUser(DbControl, Group)}
	<li>Create user: {@link #createPowerUser(DbControl, Group)}
	<li>Create guest: {@link #createGuest(DbControl, Group)}
	<li>Create fileformat for import:
	<li>Import reporters:
	</ol>
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2011-10-26 12:51:52 +0200 (Wed, 26 Oct 2011) $
*/
public class AdminTest
{

	public static boolean ok = true;
	
	/**
		Run the tests in this class.
	*/
	public static boolean run()
	{
		DbControl dc = null;
		TestUtil.write("++Entering administrator test\n");
		try
		{
			TestUtil.login("admin", "admin");
			dc = TestUtil.getDbControl();
			Group g = createGroup(dc);
			User powerUser = createPowerUser(dc, g);
			User user = createUser(dc, g);
			User guest = createGuest(dc, g);
			dc.commit();
			
			// Give permission to plugins
			dc = TestUtil.getDbControl();
			Set<Permission> USE = EnumSet.of(Permission.USE);
			Set<SharedItem> plugins = new HashSet<SharedItem>();
			plugins.add(PluginDefinition.getByClassName(dc,	PlateFlatFileImporter.class.getName()));
			plugins.add(PluginDefinition.getByClassName(dc,	ReporterMapFlatFileImporter.class.getName()));
			plugins.add(PluginDefinition.getByClassName(dc,	PrintMapFlatFileImporter.class.getName()));
			plugins.add(PluginDefinition.getByClassName(dc,	GtfReporterMapImporter.class.getName()));
			
			plugins.add(PluginDefinition.getByClassName(dc, ArrayDesignImporter.class.getName()));
			plugins.add(PluginDefinition.getByClassName(dc, ArrayBatchImporter.class.getName()));
			plugins.add(PluginDefinition.getByClassName(dc, ArraySlideImporter.class.getName()));						
			addPermissions(dc, powerUser, plugins, USE);
			dc.commit();
			
			//Configuration to import reporters
			dc = TestUtil.getDbControl();
			PluginConfiguration reporterImporter = PluginUtil.createPluginConfiguration(dc, "Reporters for project A", 
					"net.sf.basedb.plugins.ReporterFlatFileImporter", MouseData.getReporterParameters());

			PluginConfiguration genePixReporterImporter = PluginUtil.createPluginConfiguration(dc, 
					"Reporters from GenePix file",
					"net.sf.basedb.plugins.ReporterFlatFileImporter", MouseData.getGenePixReporterParameters());
			
			PluginConfiguration affyReporterImporter = PluginUtil.createPluginConfiguration(dc, 
					"Reporters from Affymetrix annotations file",
					"net.sf.basedb.plugins.ReporterFlatFileImporter", AffymetrixData.getReporterParameters());
			
			PluginConfiguration gtfReporterImporter = PluginUtil.createPluginConfiguration(dc, 
					"Reporters from GTF file",
					"net.sf.basedb.plugins.gtf.GtfReporterImporter", SequenceData.getGtfReporterParameters());
			dc.commit();
			
			//Upload file with reporters
			dc = TestUtil.getDbControl();
			File reporters = FileUtil.uploadFile(dc, "/mouse", "plates_and_reporters.mouse.v4.37k.txt", File.PLATE);
			File affyReporters = FileUtil.uploadFile(dc, "/affymetrix/annotations", "MG_U74Av2_annot.csv.tar.bz2", File.REPORTER);
			File gtfReporters = FileUtil.uploadFile(dc, "/sequencing", "UCSC_Human_hg19_RefSeqGenes.gtf.tar.bz2", File.REPORTER);
			dc.commit();
			
			//Import reporters
			PluginUtil.importReporters(reporterImporter, reporters, "Importing reporters for project A");
			PluginUtil.importReporters(affyReporterImporter, affyReporters, "Importing Affymetrix reporters for project A");
			PluginUtil.importReporters(gtfReporterImporter, gtfReporters, "Importing GTF reporters for project A");
			TestUtil.logout();
		}
		catch (Throwable t)
		{
			ok = false;
			t.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		TestUtil.write("++Leaving administrator test (" + (ok ? "OK" : "FAILED") + ")\n");
		return ok;
	}
	
	/**
		Create a common group for the users.
	*/
	public static Group createGroup(DbControl dc)
	{
		TestUtil.write("--Creating group\n");
		Quota oneGig = Util.findQuota(dc, "1 GB total");
		Group g = Group.getNew(dc);
		g.setName("Group A");
		g.setQuota(oneGig);
		dc.saveItem(g);
		return g;
	}
	
	/**
		Create user with power user rights.
	*/
	public static User createPowerUser(DbControl dc, Group quotaGroup)
	{
		TestUtil.write("--Creating power user\n");
		Role powerUser = Role.getById(dc, SystemItems.getId(Role.POWER_USER));
		User power = Util.createUser(dc, "Power user", "power", "power", quotaGroup.getQuota(), quotaGroup, 
				Collections.singleton(powerUser), null);
		Directory.createHomeDirectory(dc, power, true);
		return power;
	}

	/**
		Create user with user rights.
	*/
	public static User createUser(DbControl dc, Group quotaGroup)
	{
		TestUtil.write("--Creating regular user\n");
		Role userRole = Role.getById(dc, SystemItems.getId(Role.USER));
		User user = Util.createUser(dc, "User", "user", "user", quotaGroup.getQuota(), quotaGroup, 
				Collections.singleton(userRole), null);
		Directory.createHomeDirectory(dc, user, true);
		return user;
	}
	
	/**
		Create user with user rights.
	*/
	public static User createGuest(DbControl dc, Group quotaGroup)
	{
		TestUtil.write("--Creating guest\n");
		Role guest = Role.getById(dc, SystemItems.getId(Role.GUEST));
		Quota guestQuota = Quota.getById(dc, SystemItems.getId(Quota.DEFAULT)); //  Should be 10 MB
		
		User user = Util.createUser(dc, "Guest", "guest", "guest", guestQuota, quotaGroup, 
			Collections.singleton(guest), null);
		Directory.createHomeDirectory(dc, user, true);
		return user;
	}
	
	public static void addPermissions(DbControl dc, User user, Collection<SharedItem> items, Set<Permission> permissions)
	{
		MultiPermissions mp = new MultiPermissions(items);
		mp.setPermissions(user, permissions);
		mp.updateKeys(dc);
	}
	
}
